##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"

import logging

from .test_utils import QasmFile, get_phir_json

logger = logging.getLogger(__name__)


def test_pll_tk2_same_angle() -> None:
    """Make sure the parallelization is correct for the tk2_same_angle circuit."""
    phir = get_phir_json(QasmFile.tk2_same_angle, rebase=False)

    # Check that the op is properly formatted
    op = phir["ops"][3]
    measure = phir["ops"][5]
    assert op["qop"] == "R2XXYYZZ"

    # Check that the args are properly formatted
    assert len(op["args"]) == 2
    assert len(op["args"][0]) == len(op["args"][1]) == 2
    q01_fst = (["q", 0] in op["args"][0]) and (["q", 1] in op["args"][0])
    q01_snd = (["q", 0] in op["args"][1]) and (["q", 1] in op["args"][1])
    q23_fst = (["q", 2] in op["args"][0]) and (["q", 3] in op["args"][0])
    q23_snd = (["q", 2] in op["args"][1]) and (["q", 3] in op["args"][1])
    assert (q01_fst and q23_snd) != (q23_fst and q01_snd)

    # Check that the measure op is properly formatted
    measure_args = measure["args"]
    measure_returns = measure["returns"]
    assert len(measure_args) == len(measure_returns) == 4
    assert measure_args.index(["q", 0]) == measure_returns.index(["c", 0])
    assert measure_args.index(["q", 1]) == measure_returns.index(["c", 1])
    assert measure_args.index(["q", 2]) == measure_returns.index(["c", 2])
    assert measure_args.index(["q", 3]) == measure_returns.index(["c", 3])


def test_pll_tk2_diff_angles() -> None:
    """Make sure the parallelization is correct for the tk2_diff_angles circuit."""
    phir = get_phir_json(QasmFile.tk2_diff_angles, rebase=False)

    # Check that the qparallel block is properly formatted
    block = phir["ops"][3]
    measure = phir["ops"][5]
    assert block["block"] == "qparallel"
    assert len(block["ops"]) == 2

    # Check that the individual ops are properly formatted
    qop0, qop1 = block["ops"]
    assert qop0["qop"] == qop1["qop"] == "R2XXYYZZ"
    assert len(qop0["args"][0]) == len(qop1["args"][0]) == 2
    # Ensure the args for each op are invalid combinations, irrespecive of order
    q01_fst = (["q", 0] in qop0["args"][0]) and (["q", 1] in qop0["args"][0])
    q01_snd = (["q", 0] in qop1["args"][0]) and (["q", 1] in qop1["args"][0])
    q23_fst = (["q", 2] in qop0["args"][0]) and (["q", 3] in qop0["args"][0])
    q23_snd = (["q", 2] in qop1["args"][0]) and (["q", 3] in qop1["args"][0])
    assert (q01_fst and q23_snd) != (q23_fst and q01_snd)

    # Check that the measure op is properly foramtted
    measure_args = measure["args"]
    measure_returns = measure["returns"]
    assert len(measure_args) == len(measure_returns) == 4
    assert measure_args.index(["q", 0]) == measure_returns.index(["c", 0])
    assert measure_args.index(["q", 1]) == measure_returns.index(["c", 1])
    assert measure_args.index(["q", 2]) == measure_returns.index(["c", 2])
    assert measure_args.index(["q", 3]) == measure_returns.index(["c", 3])



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"

import json
import logging
from collections import OrderedDict
from typing import TYPE_CHECKING

import pytket.circuit as tk
from phir.model import PHIRModel

from .phirgen import PHIR_HEADER, append_cmd, arg_to_bit, get_decls, tket_gate_to_phir

if TYPE_CHECKING:
    from pytket.circuit import Circuit
    from pytket.unit_id import UnitID

    from .machine import Machine
    from .phirgen import JsonDict
    from .sharding.shard import Cost, Ordering, Shard, ShardLayer

logger = logging.getLogger(__name__)


def exec_order_preserved(
    group_exec_order: list[int], group_number: int, qubit_last_group: int
) -> bool:
    """Determine whether order is preserved when adding qubits to groups."""
    if (group_number not in group_exec_order) or (
        qubit_last_group not in group_exec_order
    ):
        return True
    if group_number <= qubit_last_group:
        return False
    # if the group that the qubit is eligible for is later in the exec_order list than
    # the last group in which it was used, it can be parallelized
    return group_exec_order.index(group_number) > group_exec_order.index(
        qubit_last_group
    )


def process_sub_commands(  # noqa: PLR0912
    sub_commands: dict["UnitID", list[tk.Command]], max_parallel_sq_gates: int
) -> dict[int, list[tk.Command]]:
    """Create parallelizable groups of sub-commands."""
    groups: dict[
        int, list[tk.Command]
    ] = {}  # be sure to order by group number into a list when returning
    qubits2groups = OrderedDict()  # the most recent group in which a qubit was used
    # group numbers for each gate are incremented by 3 so they don't overlap
    # and different gate types don't go in the same group
    # RZ gates go in (mod 3)=0 groups, R1XY gates go in (mod 3)=1 groups,
    # and all other gates will go in (mod 3)=2 groups
    rz_group_number = -3  # set to 0 when first RZ gate is assigned (-3 + 3 = 0)
    r1xy_group_number = -2  # set to 1 when first R1XY gate is assigned (-2 + 3 = 1)
    other_group_number = -1  # set to 2 when first other gate is assigned (-1 + 3 = 2)
    num_scs_per_qubit: dict[UnitID, int] = {}
    group_exec_order: list[int] = []

    for qubit, cmds in sub_commands.items():
        num_scs_per_qubit[qubit] = len(cmds)

    max_len = max(num_scs_per_qubit.values())

    for index in range(max_len):
        for qubit, cmds in sub_commands.items():
            # make sure the qubits are inserted into
            # qubits2groups in the order in which they appear
            if qubit not in qubits2groups:
                # set every qubit's group id to be a default value -4
                # prevents KeyError in check for group number
                # will get set to a valid group number the first time the qubit is used
                qubits2groups[qubit] = -4
            # check to make sure you are not accessing beyond the end of the list
            if index < num_scs_per_qubit[qubit]:
                sc = cmds[index]
                gate = sc.op.type
                match gate:
                    case tk.OpType.Rz:
                        group_number = rz_group_number
                        valid_pll_op = True
                    case tk.OpType.PhasedX:
                        group_number = r1xy_group_number
                        valid_pll_op = True
                    case _:
                        group_number = other_group_number
                        valid_pll_op = False
                # does a group exist for that gate type?
                group_available = group_number in groups
                # is the group size still under the maximum allowed parallel ops?
                group_size = len(groups[group_number]) if group_number in groups else 0
                group_not_too_large = group_size < max_parallel_sq_gates
                # is the op parallelizable (only RZ or R1XY)?
                if (
                    group_available
                    and exec_order_preserved(
                        group_exec_order, group_number, qubits2groups[qubit]
                    )
                    and group_not_too_large
                    and valid_pll_op
                ):
                    groups[group_number].append(sc)
                    qubits2groups[qubit] = group_number
                else:
                    # make a new group:
                    match gate:
                        case tk.OpType.Rz:
                            rz_group_number += 3
                            group_number = rz_group_number
                        case tk.OpType.PhasedX:
                            r1xy_group_number += 3
                            group_number = r1xy_group_number
                        case _:
                            other_group_number += 3
                            group_number = other_group_number
                    groups[group_number] = [sc]
                    group_exec_order.append(group_number)
                    qubits2groups[qubit] = group_number

    return dict(groups.items())


def groups2qops(groups: dict[int, list[tk.Command]], ops: list["JsonDict"]) -> None:  # noqa: PLR0912
    """Convert the groups of parallel ops to properly formatted PHIR."""
    for group in groups.values():
        angles2qops: dict[tuple[float, ...], JsonDict] = {}
        for qop in group:
            if not qop.op.is_gate():
                append_cmd(qop, ops)
            else:
                angles = qop.op.params
                if tuple(angles) not in angles2qops:
                    fmt_qop: JsonDict = {
                        "qop": tket_gate_to_phir[qop.op.type],
                        "angles": [angles, "pi"],
                    }
                    if len(qop.qubits) == 1:
                        fmt_qop["args"] = [arg_to_bit(qop.qubits[0])]
                    else:
                        arg = [arg_to_bit(a) for a in qop.qubits]
                        fmt_qop["args"] = [arg]
                    angles2qops[tuple(angles)] = fmt_qop
                else:
                    fmt_qop = angles2qops[tuple(angles)]
                    if len(qop.qubits) == 1:
                        fmt_qop["args"].append(arg_to_bit(qop.qubits[0]))
                    else:
                        arg = [arg_to_bit(a) for a in qop.qubits]
                        fmt_qop["args"].append(arg)
        # in the case where the qop is not a gate (a conditional for example)
        # this branch is skipped because non-gate sub-commands
        # are always the only member of their group (see process_sub_commands)
        if len(angles2qops) > 1:
            pll_block: JsonDict = {"block": "qparallel", "ops": []}
            for phir_qop in angles2qops.values():
                pll_block["ops"].append(phir_qop)
            comment = {"//": f"Parallel {tket_gate_to_phir[qop.op.type]}"}
            ops.extend((comment, pll_block))
        else:
            for phir_qop in angles2qops.values():
                if len(phir_qop["args"]) > 1:
                    comment = {"//": "Parallel " + str(qop).split(" q", maxsplit=1)[0]}
                else:
                    comment = {"//": str(qop)}
                ops.extend((comment, phir_qop))


def process_shards(
    shard_layer: "ShardLayer", max_parallel_tq_gates: int, max_parallel_sq_gates: int
) -> dict[int, list["Shard"]]:
    """Break up the shard layer into parallelizable groups."""
    groups: dict[int, list[Shard]] = {}
    types2groups: dict[
        tk.OpType, int
    ] = {}  # a dict to track the last group in which a gate type was used
    group_tracker = 0
    # parallelize shards in the same layer only
    # keeps a cap on the complexity by not having to look at
    # the relative ordering of ops in the whole circuit
    # working within layers guarantees no qubit overlap
    for shard in shard_layer:
        # the gate is a quantum op with a fixed number of arguments and
        # a known machine restriction (i.e. not Barrier, TK2, etc)
        # TODO(asa): add support for barriers
        # https://github.com/CQCL/pytket-phir/issues/55
        eligible_command_type = shard.primary_command.op.type in tket_gate_to_phir
        # a group is available for that gate type
        group_available = shard.primary_command.op.type in types2groups
        # the group size does not exceed the max number of parallel gates
        args = shard.primary_command.qubits
        num_args = len(args)
        group_not_full = False
        if group_available:
            group_number = types2groups[shard.primary_command.op.type]
            group = groups[group_number]
            tq_case = (num_args == 2) and (len(group) < max_parallel_tq_gates)  # noqa: PLR2004
            sq_case = (num_args == 1) and (len(group) < max_parallel_sq_gates)
            if tq_case or sq_case:
                group_not_full = True
        # if all 3 conditions hold, add the shard to the existing group
        if eligible_command_type and group_available and group_not_full:
            group.append(shard)
            types2groups[shard.primary_command.op.type] = group_number
        else:
            # otherwise, make a new group
            groups[group_tracker] = [shard]
            types2groups[shard.primary_command.op.type] = group_tracker
            group_tracker += 1

    groups = consolidate_sub_commands(groups)
    return dict(sorted(groups.items()))


def consolidate_sub_commands(
    groups: dict[int, list["Shard"]],
) -> dict[int, list["Shard"]]:
    """Group all the sub_commands into the first shard in the group.

    This allows maximum parallelization of sub-commands across shards.
    """
    for shards in groups.values():
        # stuckee == the shard that gets 'stuck' with all the sub-commands
        stuckee = shards[0]
        # if the group of shards only contains one element, do nothing,
        # the sub-commands are already packed into one shard
        if len(shards) > 1:
            for i in range(1, len(shards)):
                working_shard = shards[i]
                stuckee.sub_commands.update(working_shard.sub_commands)
                working_shard.sub_commands.clear()
    return groups


def format_and_add_primary_commands(
    group: list["Shard"], ops: list["JsonDict"]
) -> None:
    """Create properly formatted PHIR for parallel primary commands."""
    if len(group) == 1:
        append_cmd(group[0].primary_command, ops)
    else:
        num_angles = len(group[0].primary_command.op.params)
        # format the ops with no angles
        if num_angles == 0:
            gate_type = tket_gate_to_phir[group[0].primary_command.op.type]
            # for init, just append normally
            if gate_type == "Init":
                for shard in group:
                    append_cmd(shard.primary_command, ops)
            # for measure, format and include "returns"
            elif gate_type == "Measure":
                fmt_measure: JsonDict = {
                    "qop": "Measure",
                    "args": [],
                    "returns": [],
                }
                for shard in group:
                    pc = shard.primary_command
                    fmt_measure["args"].append(arg_to_bit(pc.args[0]))
                    fmt_measure["returns"].append(arg_to_bit(pc.bits[0]))
                ops.append(fmt_measure)
            # all other gates, treat as standard qops
            else:
                fmt_qop: JsonDict = {"qop": gate_type, "args": []}
                for shard in group:
                    pc = shard.primary_command
                    fmt_qop["args"].append(arg_to_bit(pc.args[0]))
                ops.append(fmt_qop)
        else:
            fmt_g2q: dict[int, list[tk.Command]] = {0: []}
            for shard in group:
                fmt_g2q[0].append(shard.primary_command)
            groups2qops(fmt_g2q, ops)


def get_transport_time_for_gate(gate: str, machine: "Machine") -> float:
    """Return the transport time on the machine for the given gate type."""
    match gate:
        case "RZ" | "R1XY":
            return machine.sq_time
        case "RZZ":
            return machine.tq_time
        case "Measure":
            return machine.meas_prep_time
        case "Init":
            return 0
        case _:
            logger.warning("Gate type %s not assigned a transport duration", gate)
            return 0


def adjust_phir_transport_time(ops: list["JsonDict"], machine: "Machine") -> None:
    """Analyze the generated phir and adjust the transport time."""
    adjustment = 0.0
    for op in ops:
        if "qop" in op:
            adjustment += get_transport_time_for_gate(op["qop"], machine)
        if "block" in op and op["block"] == "qparallel":
            first_op = op["ops"][0]["qop"]
            adjustment += get_transport_time_for_gate(first_op, machine)
        if "mop" in op and op["mop"] == "Transport":
            cost, units = op["duration"]
            op["duration"] = cost + adjustment, units
            adjustment = 0.0


def genphir_parallel(
    inp: list[tuple["Ordering", "ShardLayer", "Cost"]],
    circuit: "Circuit",
    machine: "Machine",
) -> str:
    """Convert a list of shards to the equivalent PHIR with parallel gating.

    Args:
        inp: list of shards
        circuit: corresponding tket Circuit
        machine: a QTM machine on which to simulate the circuit
    """
    max_parallel_tq_gates = len(machine.tq_options) // 2
    max_parallel_sq_gates = len(machine.sq_options) // 2

    phir = PHIR_HEADER
    phir["metadata"]["strict_parallelism"] = True
    ops: list[JsonDict] = []

    for _orders, shard_layer, layer_cost in inp:
        # within each shard layer, create groups of parallelizable shards
        # squash all the sub-commands into the first shard in the group
        shard_groups = process_shards(
            shard_layer, max_parallel_tq_gates, max_parallel_sq_gates
        )
        for group in shard_groups.values():
            for shard in group:
                if shard.sub_commands.values():
                    # sub-commands are always sq gates
                    subcmd_groups = process_sub_commands(
                        shard.sub_commands, max_parallel_sq_gates
                    )
                    groups2qops(subcmd_groups, ops)
            format_and_add_primary_commands(group, ops)

        ops.append(
            {
                "mop": "Transport",
                "duration": (layer_cost, "ms"),
            },
        )
    adjust_phir_transport_time(ops, machine)

    decls = get_decls(circuit.q_registers, circuit.c_registers)

    phir["ops"] = decls + ops
    PHIRModel.model_validate(phir)
    return json.dumps(phir)



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"

import logging

from .test_utils import QasmFile, get_phir_json

logger = logging.getLogger(__name__)


def test_parallelization() -> None:
    """Make sure the parallelization is happening properly for the test circuit."""
    phir = get_phir_json(QasmFile.parallelization_test, rebase=True)

    # Make sure The parallel RZ and R1XY gates have the correct arguments
    parallel_rz1 = phir["ops"][3]
    assert parallel_rz1["qop"] == "RZ"
    qubits = [["q", 0], ["q", 1], ["q", 2], ["q", 3]]
    for qubit in qubits:
        assert qubit in parallel_rz1["args"]
    parallel_phasedx = phir["ops"][5]
    assert parallel_phasedx["qop"] == "R1XY"
    for qubit in qubits:
        assert qubit in parallel_phasedx["args"]

    # Make sure the parallel block is properly formatted
    block = phir["ops"][7]
    assert block["block"] == "qparallel"
    assert len(block["ops"]) == 2
    qop0 = block["ops"][0]
    qop1 = block["ops"][1]
    assert qop0["qop"] == qop1["qop"] == "RZZ"

    # Make sure the ops within the parallel block have the correct arguments
    assert len(qop0["args"][0]) == len(qop1["args"][0]) == 2
    q01_fst = (["q", 0] in qop0["args"][0]) and (["q", 1] in qop0["args"][0])
    q01_snd = (["q", 0] in qop1["args"][0]) and (["q", 1] in qop1["args"][0])
    q23_fst = (["q", 2] in qop0["args"][0]) and (["q", 3] in qop0["args"][0])
    q23_snd = (["q", 2] in qop1["args"][0]) and (["q", 3] in qop1["args"][0])
    assert (q01_fst and q23_snd) != (q23_fst and q01_snd)

    # Make sure the measure op is properly formatted
    measure = phir["ops"][9]
    measure_args = measure["args"]
    measure_returns = measure["returns"]
    assert len(measure_args) == len(measure_returns) == 4
    assert measure_args.index(["q", 0]) == measure_returns.index(["c", 0])
    assert measure_args.index(["q", 1]) == measure_returns.index(["c", 1])
    assert measure_args.index(["q", 2]) == measure_returns.index(["c", 2])
    assert measure_args.index(["q", 3]) == measure_returns.index(["c", 3])


def test_parallel_subcommand_relative_ordering() -> None:
    """Make sure the proper relative ordering of sub-commands is preserved."""
    phir = get_phir_json(QasmFile.rxrz, rebase=True)
    # make sure it is ordered like the qasm file
    ops = phir["ops"]
    sc1 = ops[3]
    sc2 = ops[5]
    sc3 = ops[7]
    sc4 = ops[9]
    assert sc1["qop"] == "RZ"
    assert sc1["angles"] == [[0.5], "pi"]
    assert sc2["qop"] == "R1XY"
    assert sc2["angles"] == [[3.5, 0.0], "pi"]
    assert sc3["qop"] == "R1XY"
    assert sc3["angles"] == [[0.5, 0.0], "pi"]
    assert sc4["qop"] == "RZ"
    assert sc4["angles"] == [[3.5], "pi"]


def test_single_qubit_circuit_with_parallel() -> None:
    """Make sure there are no parallel blocks present in the 1qubit circuit."""
    phir_with_parallel_phirgen = get_phir_json(
        QasmFile.single_qubit_parallel_test, rebase=True
    )
    phir_with_standard_phirgen = get_phir_json(
        QasmFile.single_qubit_parallel_test, rebase=False
    )
    assert len(phir_with_parallel_phirgen) == len(phir_with_standard_phirgen)
    # since the rebasing converts to the native gate set,
    # the names and angle foramts of the qops will not match.
    # for example Ry gets converted to R1XY
    # compare angles and args instead

    ops = (3, 5, 7, 12, 14, 16)

    assert phir_with_parallel_phirgen["ops"][9] == {
        "meta": "barrier",
        "args": [["q", 0]],
    }

    for i, qop in zip(ops, ("R1XY", "RZ", "R1XY", "R1XY", "RZ", "R1XY"), strict=True):
        assert phir_with_parallel_phirgen["ops"][i]["qop"] == qop

    for i, qop in zip(ops, ("RY", "RZ", "RY", "RY", "RZ", "RY"), strict=True):
        assert phir_with_standard_phirgen["ops"][i]["qop"] == qop

    for i in ops:
        assert (
            phir_with_parallel_phirgen["ops"][i]["args"]
            == phir_with_standard_phirgen["ops"][i]["args"]
        )


def test_three_qubit_rz_exec_order_preserved() -> None:
    """Test that the order of gating is preserved in a 3 qubit circuit with RZ gates."""
    phir = get_phir_json(QasmFile.rz_exec_order_three_qubits, rebase=True)
    # verify that the parallel RZ gates are executed before the second R1XY gate
    assert phir["ops"][8] == {
        "qop": "R1XY",
        "angles": [[0.5, 0.5], "pi"],
        "args": [["q", 0]],
    }
    assert phir["ops"][10] == {
        "block": "qparallel",
        "ops": [
            {"qop": "RZ", "angles": [[0.5], "pi"], "args": [["q", 1]]},
            {"qop": "RZ", "angles": [[1.5], "pi"], "args": [["q", 2]]},
        ],
    }
    assert phir["ops"][12] == {
        "qop": "R1XY",
        "angles": [[0.5, 0.5], "pi"],
        "args": [["q", 2]],
    }


def test_two_qubit_exec_order_preserved() -> None:
    """Test that the order of gating in preserved in a 2 qubit circuit."""
    phir = get_phir_json(QasmFile.exec_order_two_qubits, rebase=True)

    assert phir["ops"][16] == {"meta": "barrier", "args": [["q", 0], ["q", 1]]}
    # for this test, verify that the RX gates are NOT parallelized
    # in the following section of the qasm file (lines 11-13):
    # rx(3.5*pi) q[0];
    assert phir["ops"][19] == {
        "qop": "R1XY",
        "angles": [[3.5, 0.0], "pi"],
        "args": [["q", 0]],
    }
    # rz(3.5*pi) q[1];
    assert phir["ops"][21] == {"qop": "RZ", "angles": [[3.5], "pi"], "args": [["q", 1]]}
    # rx(1.0*pi) q[1];
    assert phir["ops"][25] == {
        "qop": "R1XY",
        "angles": [[1.0, 0.0], "pi"],
        "args": [["q", 1]],
    }


def test_group_ordering() -> None:
    """Test that groups are in the right order when the group number can decrement."""
    phir = get_phir_json(QasmFile.group_ordering, rebase=True)
    block = phir["ops"][3]
    assert block["block"] == "qparallel"
    assert block["ops"][0] == {
        "qop": "R1XY",
        "angles": [[0.5, 0.0], "pi"],
        "args": [["q", 0]],
    }
    assert block["ops"][1] == {
        "qop": "R1XY",
        "angles": [[3.5, 0.5], "pi"],
        "args": [["q", 1]],
    }
    assert phir["ops"][5] == {
        "qop": "R1XY",
        "angles": [[0.5, 0.0], "pi"],
        "args": [["q", 0]],
    }
    assert phir["ops"][7] == {"qop": "RZ", "angles": [[1.0], "pi"], "args": [["q", 1]]}
    assert phir["ops"][9] == {
        "qop": "R1XY",
        "angles": [[3.5, 0.5], "pi"],
        "args": [["q", 1]],
    }



##############################################################################
#
# Copyright (c) 2023-2024 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"

import json
import logging
import sys
from collections import deque
from copy import deepcopy
from importlib.metadata import version
from typing import TYPE_CHECKING, Any, TypeAlias

from pytket.qasm.qasm import QASMUnsupportedError

if sys.version_info >= (3, 11):
    from typing import assert_never
else:
    from typing_extensions import assert_never

import pytket
import pytket.circuit as tk
from phir.model import PHIRModel
from pytket.circuit import ClBitVar, ClExpr, ClOp, ClRegVar
from pytket.circuit.clexpr import has_reg_output
from pytket.circuit.logic_exp import (
    BitLogicExp,
    BitWiseOp,
    Constant,
    LogicExp,
    RegLogicExp,
    RegWiseOp,
)
from pytket.unit_id import Bit as tkBit
from pytket.unit_id import BitRegister, QubitRegister

if TYPE_CHECKING:
    from collections.abc import Sequence

    from pytket.circuit import Circuit, WiredClExpr
    from pytket.unit_id import UnitID

    from .sharding.shard import Cost, Ordering, ShardLayer

logger = logging.getLogger(__name__)

JsonDict: TypeAlias = dict[str, Any]
PHIR_HEADER: JsonDict = {
    "format": "PHIR/JSON",
    "version": "0.1.0",
    "metadata": {"source": f"pytket-phir v{version('pytket-phir').split('+')[0]}"},
}

WASM_WORDSIZE = 32
WORDSIZE = 64 if pytket.__dict__.get("bit_width_64", False) else 32
UINTMAX = 2**WORDSIZE - 1

Var: TypeAlias = str
Bit: TypeAlias = list[Var | int]  # e.g. [c, 0] for c[0]

tket_gate_to_phir = {
    tk.OpType.Reset:    "Init",
    tk.OpType.Measure:  "Measure",
    tk.OpType.noop:     "I",

    tk.OpType.CX:       "CX",
    tk.OpType.CY:       "CY",
    tk.OpType.CZ:       "CZ",
    tk.OpType.H:        "H",
    tk.OpType.PhasedX:  "R1XY",
    tk.OpType.Rx:       "RX",
    tk.OpType.Ry:       "RY",
    tk.OpType.Rz:       "RZ",
    tk.OpType.S:        "SZ",
    tk.OpType.Sdg:      "SZdg",
    tk.OpType.SWAP:     "SWAP",
    tk.OpType.SX:       "SX",
    tk.OpType.SXdg:     "SXdg",
    tk.OpType.T:        "T",
    tk.OpType.Tdg:      "Tdg",
    tk.OpType.TK2:      "R2XXYYZZ",
    tk.OpType.U1:       "RZ",
    tk.OpType.V:        "SX",
    tk.OpType.Vdg:      "SXdg",
    tk.OpType.X:        "X",
    tk.OpType.XXPhase:  "RXX",
    tk.OpType.Y:        "Y",
    tk.OpType.YYPhase:  "RYY",
    tk.OpType.Z:        "Z",
    tk.OpType.ZZMax:    "SZZ",
    tk.OpType.ZZPhase:  "RZZ",
}  # fmt: skip


def arg_to_bit(arg: "UnitID") -> Bit:
    """Convert tket arg to Bit."""
    return [arg.reg_name, arg.index[0]]


def assign_cop(
    lhs: list[Var] | list[Bit], rhs: "Sequence[Var | int | JsonDict | Bit]"
) -> JsonDict:
    """PHIR for classical assign operation."""
    return {
        "cop": "=",
        "returns": lhs,
        "args": rhs,
    }


def classical_op(exp: LogicExp, *, bitwise: bool = False) -> JsonDict | int:  # noqa: PLR0912, PLR0915
    """PHIR for classical register operations."""
    match exp.op:
        # Nullary
        case BitWiseOp.ZERO:
            return 0
        case BitWiseOp.ONE:
            return 1
        # Bitwise
        case RegWiseOp.AND | BitWiseOp.AND:
            cop = "&"
        case RegWiseOp.OR | BitWiseOp.OR:
            cop = "|"
        case RegWiseOp.XOR | BitWiseOp.XOR:
            cop = "^"
        case RegWiseOp.NOT | BitWiseOp.NOT:
            cop = "~"
        case RegWiseOp.LSH:
            cop = "<<"
        case RegWiseOp.RSH:
            cop = ">>"
        # Comparison
        case RegWiseOp.EQ | BitWiseOp.EQ:
            cop = "=="
        case RegWiseOp.NEQ | BitWiseOp.NEQ:
            cop = "!="
        case RegWiseOp.LT:
            cop = "<"
        case RegWiseOp.GT:
            cop = ">"
        case RegWiseOp.LEQ:
            cop = "<="
        case RegWiseOp.GEQ:
            cop = ">="
        # Arithmetic
        case RegWiseOp.ADD:
            cop = "+"
        case RegWiseOp.SUB | RegWiseOp.NEG:
            cop = "-"
        case RegWiseOp.MUL:
            cop = "*"
        case RegWiseOp.DIV:
            cop = "/"
        case RegWiseOp.POW:
            cop = "**"
        case _:
            assert_never(exp.op)

    args: list[JsonDict | Var | Constant | Bit] = []
    for arg in exp.args:
        match arg:
            case BitLogicExp():
                args.append(classical_op(arg, bitwise=True))
            case LogicExp():
                args.append(classical_op(arg))
            case BitRegister():
                if bitwise:
                    raise TypeError
                args.append(arg.name)
            case Constant():
                if bitwise and (arg < 0 or arg > 1):
                    raise ValueError
                args.append(arg)
            case tkBit():
                if bitwise:
                    args.append(arg_to_bit(arg))
                else:
                    args.append(arg.reg_name)
            case _:
                assert_never(arg)
    return {
        "cop": cop,
        "args": args,
    }


def convert_gate(op: tk.Op, cmd: tk.Command) -> JsonDict | None:
    """Return PHIR dict for a tket gate op."""
    try:
        gate = tket_gate_to_phir[op.type]
    except KeyError:
        if op.type == tk.OpType.Phase:
            # ignore global phase
            return {"mop": "Skip"}
        logging.exception("Gate %s unsupported by PHIR", op.get_name())
        raise

    angles = (op.params, "pi") if op.params else None
    qop: JsonDict
    match gate:
        case "Measure":
            qop = {
                "qop": gate,
                "returns": [arg_to_bit(cmd.bits[0])],
                "args": [arg_to_bit(cmd.qubits[0])],
            }
        case ("CX"
            | "CY"
            | "CZ"
            | "RXX"
            | "RYY"
            | "RZZ"
            | "R2XXYYZZ"
            | "SXX"
            | "SXXdg"
            | "SYY"
            | "SYYdg"
            | "SZZ"
            | "SZZdg"
            | "SWAP"
        ):  # two-qubit gates  # fmt: skip
            qop = {
                "qop": gate,
                "angles": angles,
                "args": [[arg_to_bit(cmd.qubits[0]), arg_to_bit(cmd.qubits[1])]],
            }
        case _:  # single-qubit gates
            qop = {
                "qop": gate,
                "angles": angles,
                "args": [arg_to_bit(cmd.qubits[0])],
            }
    return qop


def cop_from_op_name(op_name: str) -> str:
    """Get PHIR classical op name from pytket op name."""
    match op_name:
        case "AND":
            cop = "&"
        case "OR":
            cop = "|"
        case "XOR":
            cop = "^"
        case "NOT":
            cop = "~"
        case name:
            raise NotImplementedError(name)
    return cop


def convert_classicalevalop(op: tk.ClassicalEvalOp, cmd: tk.Command) -> JsonDict | None:  # noqa: PLR0912
    """Return PHIR dict for a pytket ClassicalEvalOp."""
    # Exclude conditional bits from args
    args = cmd.args[cmd.op.width :] if isinstance(cmd.op, tk.Conditional) else cmd.args
    out: JsonDict | None = None
    match op:
        case tk.CopyBitsOp():
            if len(cmd.bits) != len(args) // 2:
                msg = "LHS and RHS lengths mismatch for CopyBits"
                raise TypeError(msg)
            out = assign_cop(
                [arg_to_bit(bit) for bit in cmd.bits],
                [arg_to_bit(args[i]) for i in range(len(args) // 2)],
            )
        case tk.SetBitsOp():
            if len(cmd.bits) != len(op.values):
                logger.error("LHS and RHS lengths mismatch for classical assignment")
                raise ValueError
            out = assign_cop(
                [arg_to_bit(bit) for bit in cmd.bits], list(map(int, op.values))
            )
        case tk.RangePredicateOp():  # where the condition is a range
            cond: JsonDict
            match op.lower, op.upper:
                case l, u if l == u:
                    cond = {
                        "cop": "==",
                        "args": [args[0].reg_name, u],
                    }
                case l, u if u == UINTMAX:
                    cond = {
                        "cop": ">=",
                        "args": [args[0].reg_name, l],
                    }
                case 0, u:
                    cond = {
                        "cop": "<=",
                        "args": [args[0].reg_name, u],
                    }
            out = {
                "block": "if",
                "condition": cond,
                "true_branch": [assign_cop([arg_to_bit(cmd.bits[0])], [1])],
            }
        case tk.MultiBitOp():
            if len(args) % len(cmd.bits) != 0:
                msg = "Input bit- and output bit lengths mismatch."
                raise TypeError(msg)

            cop = cop_from_op_name(op.basic_op.get_name())
            is_explicit = op.basic_op.type == tk.OpType.ExplicitPredicate

            # determine number of register operands involved in the operation
            operand_count = len(args) // len(cmd.bits) - is_explicit

            iters = [iter(args)] * (operand_count + is_explicit)
            iter2 = deepcopy(iters)

            # Columns of expressions, e.g.,
            #   AND (*2) a[0], b[0], c[0]
            #          , a[1], b[1], c[1]
            #   would be [(a[0], a[1]), (b[0], b[1]), (c[0], c[1])]
            # and AND (*2) a[0], a[1], b[0]
            #            , b[1], c[0], c[1]
            #   would be [(a[0], b[1]), (a[1], c[0]), (b[0], c[1])]
            cols = zip(*zip(*iters, strict=True), strict=True)

            if all(
                all(col[0].reg_name == bit.reg_name for bit in col) for col in cols
            ):  # expression can be applied register-wise
                out = assign_cop(
                    [cmd.bits[0].reg_name],
                    [
                        {
                            "cop": cop,
                            "args": [arg.reg_name for arg in args[:operand_count]],
                        }
                    ],
                )
            else:  # apply a sequence of bit-wise ops
                exps = zip(*iter2, strict=True)
                out = {
                    "block": "sequence",
                    "ops": [
                        assign_cop(
                            [arg_to_bit(bit)],
                            [
                                {
                                    "cop": cop,
                                    "args": [
                                        arg_to_bit(arg) for arg in exp[:operand_count]
                                    ],
                                }
                            ],
                        )
                        for bit, exp in zip(cmd.bits, exps, strict=True)
                    ],
                }
        case _:
            raise NotImplementedError(op)

    return out


def multi_bit_condition(args: "list[UnitID]", value: int) -> JsonDict:
    """Construct bitwise condition."""
    min_args = 2
    if len(args) < min_args:
        msg = f"multi_bit_condition requires at least {min_args} arguments"
        raise TypeError(msg)

    def nested_cop(cop: str, args: "deque[UnitID]", val_bits: deque[int]) -> JsonDict:
        if len(args) == min_args:
            return {
                "cop": cop,
                "args": [
                    {"cop": "==", "args": [arg_to_bit(args.popleft()), val_bits.pop()]},
                    {"cop": "==", "args": [arg_to_bit(args.popleft()), val_bits.pop()]},
                ],
            }
        return {
            "cop": cop,
            "args": [
                {"cop": "==", "args": [arg_to_bit(args.popleft()), val_bits.pop()]},
                nested_cop(cop, args, val_bits),
            ],
        }

    return nested_cop("&", deque(args), deque(map(int, f"{value:0{len(args)}b}")))


def get_cop_from_op(op: ClOp) -> str | int:  # noqa: PLR0912
    """Get PHIR classical op name from ClOp."""
    cop: str | int
    match op:
        case ClOp.BitZero | ClOp.RegZero:
            cop = 0
        case ClOp.BitOne:
            cop = 1
        case ClOp.RegOne:
            cop = -1
        case ClOp.BitAnd | ClOp.RegAnd:
            cop = "&"
        case ClOp.BitOr | ClOp.RegOr:
            cop = "|"
        case ClOp.BitXor | ClOp.RegXor:
            cop = "^"
        case ClOp.BitNot | ClOp.RegNot:
            cop = "~"
        case ClOp.RegLsh:
            cop = "<<"
        case ClOp.RegRsh:
            cop = ">>"
        case ClOp.BitEq | ClOp.RegEq:
            cop = "=="
        case ClOp.BitNeq | ClOp.RegNeq:
            cop = "!="
        case ClOp.RegLt:
            cop = "<"
        case ClOp.RegGt:
            cop = ">"
        case ClOp.RegLeq:
            cop = "<="
        case ClOp.RegGeq:
            cop = ">="
        case ClOp.RegAdd:
            cop = "+"
        case ClOp.RegSub:
            cop = "-"
        case ClOp.RegMul:
            cop = "*"
        case ClOp.RegDiv:
            cop = "/"
        case ClOp.RegPow:
            cop = "**"
        case _:
            logging.exception("Classical operation %s unsupported by PHIR", str(op))
            raise NotImplementedError(op)
    return cop


def phir_from_clexpr_arg(
    expr_arg: int | ClBitVar | ClRegVar | ClExpr,
    bit_posn: dict[int, int],
    reg_posn: dict[int, list[int]],
    bits: list[tkBit],
) -> int | str | list[str | int] | JsonDict:
    """Return PHIR dict for a ClExpr."""
    match expr_arg:
        case int():
            return expr_arg
        case ClBitVar():
            bit: tkBit = bits[bit_posn[expr_arg.index]]
            return arg_to_bit(bit)
        case ClRegVar():
            bits_in_reg = [bits[i] for i in reg_posn[expr_arg.index]]
            reg_size = len(bits_in_reg)
            if reg_size == 0:
                logging.exception("Register variable with no bits")
            reg_name = bits_in_reg[0].reg_name
            if any(bit.reg_name != reg_name for bit in bits_in_reg) or any(
                bit.index[0] != i for i, bit in enumerate(bits_in_reg)
            ):
                logging.exception("Register variable not aligned with any register")
            return reg_name
    assert isinstance(expr_arg, ClExpr)  # noqa: S101

    cop = get_cop_from_op(expr_arg.op)
    if isinstance(cop, int):
        return cop
    args = [
        phir_from_clexpr_arg(arg, bit_posn, reg_posn, bits) for arg in expr_arg.args
    ]
    return {"cop": cop, "args": args}


def convert_subcmd(op: tk.Op, cmd: tk.Command) -> JsonDict | None:  # noqa: PLR0912
    """Return PHIR dict given a tket op and its arguments."""
    if op.is_gate():
        return convert_gate(op, cmd)

    out: JsonDict | None = None
    rhs: list[int | str | list[str | int] | JsonDict] = []
    match op:  # non-quantum op
        case tk.Conditional():
            out = {
                "block": "if",
                "condition": {"cop": "==", "args": [arg_to_bit(cmd.args[0]), op.value]}
                if op.width == 1
                else multi_bit_condition(cmd.args[: op.width], op.value),
                "true_branch": [convert_subcmd(op.op, cmd)],
            }

        case tk.BarrierOp():
            if op.data:
                # See https://github.com/CQCL/tket/blob/0ec603986821d994caa3a0fb9c4640e5bc6c0a24/pytket/pytket/qasm/qasm.py#L419-L459
                match op.data[0:5]:
                    case "sleep":
                        duration = op.data.removeprefix("sleep(").removesuffix(")")
                        out = {
                            "mop": "Idle",
                            "args": [arg_to_bit(qbit) for qbit in cmd.qubits],
                            "duration": (float(duration), "s"),
                        }
                    case "order" | "group":
                        raise NotImplementedError(op.data)
                    case _:
                        raise TypeError(op.data)
            else:
                out = {
                    "meta": "barrier",
                    "args": [arg_to_bit(qbit) for qbit in cmd.qubits],
                }

        case tk.ClassicalExpBox():
            exp = op.get_exp()
            match exp:
                case BitLogicExp():
                    rhs = [classical_op(exp, bitwise=True)]
                    out = assign_cop([arg_to_bit(cmd.bits[0])], rhs)
                case RegLogicExp():
                    rhs = [classical_op(exp)]
                    out = assign_cop([cmd.bits[0].reg_name], rhs)

        case tk.ClExprOp():
            wexpr: WiredClExpr = op.expr
            expr: ClExpr = wexpr.expr
            bit_posn: dict[int, int] = wexpr.bit_posn
            reg_posn: dict[int, list[int]] = wexpr.reg_posn
            output_posn: list[int] = wexpr.output_posn
            cmd_args: list[tkBit] = cmd.bits

            # TODO(AE): Check that all ClExprOps in the circuit are register-aligned
            # (i.e. that each register variable, and the register output if applicable,
            # comprises bits that constitute a complete register in the correct order).
            # https://github.com/CQCL/tket/issues/1644

            rhs = [phir_from_clexpr_arg(expr, bit_posn, reg_posn, cmd_args)]
            if has_reg_output(expr.op):
                return assign_cop([cmd_args[output_posn[0]].reg_name], rhs)
            return assign_cop([arg_to_bit(cmd_args[output_posn[0]])], rhs)

        case tk.ClassicalEvalOp():
            return convert_classicalevalop(op, cmd)

        case tk.WASMOp():
            return create_wasm_op(cmd, op)

        case _:
            # Exclude conditional bits from args
            args = (
                cmd.args[cmd.op.width :]
                if isinstance(cmd.op, tk.Conditional)
                else cmd.args
            )
            match op.type:
                case tk.OpType.ExplicitPredicate | tk.OpType.ExplicitModifier:
                    # exclude output bit when not modifying in place
                    args = args[:-1] if op.type == tk.OpType.ExplicitPredicate else args
                    out = assign_cop(
                        [arg_to_bit(cmd.bits[0])],
                        [
                            {
                                "cop": cop_from_op_name(op.get_name()),
                                "args": [arg_to_bit(arg) for arg in args],
                            }
                        ],
                    )
                case _:
                    raise NotImplementedError(op.type)

    return out


def append_cmd(cmd: tk.Command, ops: list[JsonDict]) -> None:
    """Convert a pytket command to a PHIR command and append to `ops`.

    Args:
        cmd: pytket command obtained from pytket-phir
        ops: the list of ops to append to
    """
    ops.append({"//": make_comment_text(cmd, cmd.op)})
    op: JsonDict | None = convert_subcmd(cmd.op, cmd)
    if op:
        ops.append(op)


def create_wasm_op(cmd: tk.Command, wasm_op: tk.WASMOp) -> JsonDict:
    """Creates a PHIR operation for a WASM command."""
    args, returns = extract_wasm_args_and_returns(cmd, wasm_op)
    op = {
        "cop": "ffcall",
        "function": wasm_op.func_name,
        "args": args,
        "metadata": {
            "ff_object": f"WASM module uid: {wasm_op.wasm_uid}",
        },
    }
    if cmd.bits:
        op["returns"] = returns

    return op


def extract_wasm_args_and_returns(
    command: tk.Command, op: tk.WASMOp
) -> tuple[list[str], list[str]]:
    """Extract the wasm args and return values as whole register names."""
    # This slice removes the extra `_w` cregs (wires) that are not part of the
    # circuit and the output args, which are appended after the input args
    slice_index = op.num_w + sum(op.output_widths)
    only_args = command.args[:-slice_index]
    # Eliminate conditional bits from the front of the args
    input_args = only_args[len(only_args) - op.n_inputs :]
    if any(arg.index[0] >= WASM_WORDSIZE for arg in input_args):
        msg = "WASM support is limited to at most 32-bit registers"
        raise QASMUnsupportedError(msg)
    return (
        dedupe_bits_to_registers(input_args),
        dedupe_bits_to_registers(command.bits),
    )


def dedupe_bits_to_registers(bits: "Sequence[UnitID]") -> list[str]:
    """Dedupes a list of bits to their registers, keeping order intact."""
    return list(dict.fromkeys([bit.reg_name for bit in bits]))


def make_comment_text(cmd: tk.Command, op: tk.Op) -> str:
    """Converts a command + op to the PHIR comment spec."""
    comment = str(cmd)
    match op:
        case tk.Conditional():
            conditional_text = str(cmd)
            cleaned = (
                conditional_text[: conditional_text.find("THEN") + 5]
                if isinstance(op.op, tk.WASMOp)
                else ""
            )
            comment = f"{cleaned}{make_comment_text(cmd, op.op)}"

        case tk.WASMOp():
            args, returns = extract_wasm_args_and_returns(cmd, op)
            comment = f"WASM_function='{op.func_name}' args={args} returns={returns};"

        case tk.BarrierOp():
            comment = op.data + " " + str(cmd.args[0]) + ";" if op.data else str(cmd)

        case tk.ClassicalExpBox():
            exp = op.get_exp()
            match exp:
                case BitLogicExp():
                    comment = str(cmd.bits[0]) + " = " + str(op.get_exp())
                case RegLogicExp():
                    comment = str(cmd.bits[0].reg_name) + " = " + str(op.get_exp())

        case tk.ClExprOp():
            comment = (
                str(cmd).split(";")[0] + " of the form " + str(op.expr).split(" [")[0]
            )

    return comment


def get_decls(
    qregs: list[QubitRegister], cregs: list[BitRegister]
) -> list[dict[str, str | int]]:
    """Get PHIR declarations for qubits and classical variables."""
    decls: list[dict[str, str | int]] = [
        {
            "data": "qvar_define",
            "data_type": "qubits",
            "variable": qreg.name,
            "size": qreg.size,
        }
        for qreg in qregs
    ]

    decls += [
        {
            "data": "cvar_define",
            "data_type": f"i{WORDSIZE}",
            "variable": creg.name,
            "size": creg.size,
        }
        for creg in cregs
        if creg.name != "_w"
    ]

    return decls


def genphir(
    inp: list[tuple["Ordering", "ShardLayer", "Cost"]],
    circuit: "Circuit",
    *,
    machine_ops: bool = True,
) -> str:
    """Convert a list of shards to the equivalent PHIR.

    Args:
        inp: list of shards
        circuit: corresponding tket Circuit
        machine_ops: whether to include machine ops
    """
    phir = PHIR_HEADER
    ops: list[JsonDict] = []

    for _orders, shard_layer, layer_cost in inp:
        for shard in shard_layer:
            for sub_commands in shard.sub_commands.values():
                for sc in sub_commands:
                    append_cmd(sc, ops)
            append_cmd(shard.primary_command, ops)
        if machine_ops:
            ops.append(
                {
                    "mop": "Transport",
                    "duration": (layer_cost, "ms"),
                },
            )

    decls = get_decls(circuit.q_registers, circuit.c_registers)

    phir["ops"] = decls + ops
    PHIRModel.model_validate(phir)
    return json.dumps(phir)



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from typing import TYPE_CHECKING, TypeAlias

if TYPE_CHECKING:
    from pytket.unit_id import UnitID

    from .shard import Shard, ShardLayer

Layer: TypeAlias = list[list[int]]


def parse_shards_naive(
    shards: set["Shard"],
) -> tuple[list[Layer], list["ShardLayer"]]:
    """Parse a set of shards and return a circuit representation for placement."""
    layers: list[Layer] = []
    shards_in_layer: list[ShardLayer] = []
    scheduled: set[int] = set()
    num_shards: int = len(shards)
    qid_count: int = 0
    qubits2ids: dict[UnitID, int] = {}

    while len(scheduled) < num_shards:
        layer: Layer = []

        # Iterate the shards, looking for shards whose dependencies have been
        # satisfied, or initially, shards with no dependencies
        to_schedule: ShardLayer = [
            s for s in shards if s.depends_upon.issubset(scheduled)
        ]
        shards_in_layer.append(to_schedule)
        shards.difference_update(to_schedule)

        for shard in to_schedule:
            op: list[int] = []
            # if there are more than 2 qubits used, treat them all as parallel sq ops
            # one qubit will just be a single sq op
            # 3 or more will be 3 or more parallel sq ops
            # when iterating through qubits,
            # map all the qubits to a unique id to prevent duplicates in placement
            if len(shard.qubits_used) != 2:  # noqa: PLR2004
                for qubit in shard.qubits_used:
                    qid, qid_count = get_qid(qubit, qubits2ids, qid_count)
                    op = [qid]
                    layer.append(op)
            else:
                for qubit in shard.qubits_used:
                    qid, qid_count = get_qid(qubit, qubits2ids, qid_count)
                    op.append(qid)
                layer.append(op)

            scheduled.add(shard.ID)

        layers.append(layer)

    return layers, shards_in_layer


def get_qid(
    qubit: "UnitID", qubits2ids: dict["UnitID", int], qid_count: int
) -> tuple[int, int]:
    """Get qubit ID even if it is missing in the dict."""
    qid = qubits2ids.setdefault(qubit, qid_count)
    if qid == qid_count:
        qid_count += 1
    return qid, qid_count



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

import logging

from pytket.circuit import Circuit, Command, Conditional, Op, OpType
from pytket.unit_id import Bit, Qubit, UnitID

from .shard import Shard

NOT_IMPLEMENTED_OP_TYPES = [OpType.CircBox]

SHARD_TRIGGER_OP_TYPES = [
    OpType.Measure,
    OpType.Reset,
    OpType.Barrier,
    OpType.SetBits,
    OpType.ClassicalExpBox,  # some classical operations are rolled up into a box
    OpType.ClExpr,
    OpType.RangePredicate,
    OpType.ExplicitPredicate,
    OpType.ExplicitModifier,
    OpType.MultiBit,
    OpType.CopyBits,
    OpType.WASM,
]

logger = logging.getLogger(__name__)


class Sharder:
    """The Sharder class.

    Responsible for taking in a circuit in TKET representation
    and converting it into shards that can be subsequently handled in the
    compilation pipeline.
    """

    def __init__(self, circuit: Circuit) -> None:
        """Create Sharder object.

        Args:
            circuit: tket Circuit
        """
        self._circuit = circuit
        self._pending_commands: dict[UnitID, list[Command]] = {}
        self._shards: list[Shard] = []
        # These dictionaries map qubits/bits to the last shard that modified them
        self._qubit_touched_by: dict[UnitID, int] = {}
        self._bit_read_by: dict[UnitID, int] = {}
        self._bit_written_by: dict[UnitID, int] = {}

        logger.debug("Sharder created for circuit %s", self._circuit)

    def shard(self) -> list[Shard]:
        """Performs sharding algorithm on the circuit the Sharder was initialized with.

        Returns:
            list of Shards needed to schedule
        """
        logger.debug("Sharding beginning")
        commands = self._circuit.get_commands()

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("All commands:")
            for command in commands:
                logger.debug(command)

        for command in commands:
            self._process_command(command)
        self._cleanup_remaining_commands()

        logger.debug("Shard output:")
        for shard in self._shards:
            logger.debug(shard)
        return self._shards

    def _process_command(self, command: Command) -> None:
        """Handles a command per the type and the extant context within the Sharder.

        Args:
            command: tket command (operation, bits, etc)
        """
        logger.debug(
            "Processing command: %s of type %s with args: %s",
            command.op,
            command.op.type,
            command.args,
        )
        if command.op.type in NOT_IMPLEMENTED_OP_TYPES:
            msg = f"OpType {command.op.type} not supported!"
            raise NotImplementedError(msg)

        if Sharder._is_command_global_phase(command):
            logger.debug("Ignoring global Phase gate")
            return

        if Sharder.should_op_create_shard(command.op):
            self._build_shard(command)
        else:
            self._add_pending_sub_command(command)

    def _build_shard(self, command: Command) -> None:
        """Builds a shard.

        Creates a Shard object given the extant sharding context and the schedulable
        Command object passed in, and appends it to the Shard list.

        Args:
            command: tket command (operation, bits, etc)
        """
        logger.debug("Building shard for command: %s", command)
        # Rollup any sub commands (SQ gates) that interact with the same qubits
        sub_commands: dict[UnitID, list[Command]] = {}
        for key in (
            key for key in list(self._pending_commands) if key in command.qubits
        ):
            sub_commands[key] = self._pending_commands.pop(key)

        all_commands = [command]
        for sub_command_list in sub_commands.values():
            all_commands.extend(sub_command_list)

        logger.debug("All shard commands: %s", all_commands)
        qubits_used = set(command.qubits)
        bits_written = set(command.bits)
        bits_read: set[Bit] = set()

        for sub_command in all_commands:
            bits_written.update(sub_command.bits)
            bits_read.update(
                set(filter(lambda x: isinstance(x, Bit), sub_command.args)),  # type: ignore [misc, arg-type]
            )

        # Handle dependency calculations
        depends_upon = self._resolve_shard_dependencies(
            qubits_used, bits_written, bits_read
        )

        shard = Shard(
            command,
            sub_commands,
            qubits_used,
            bits_written,
            bits_read,
            depends_upon,
        )

        self._mark_dependencies(shard)

        self._shards.append(shard)
        logger.debug("Appended shard: %s", shard)

    def _resolve_shard_dependencies(
        self, qubits: set[Qubit], bits_written: set[Bit], bits_read: set[Bit]
    ) -> set[int]:
        """Finds the dependent shards for a given shard.

        This involves checking for qubit interaction and classical hazards of
        various types.

        Args:
            shard: Shard to run dependency calculation on
            qubits: Set of all qubits interacted with in the command/sub-commands
            bits_written: Classical bits the command/sub-commands write to
            bits_read: Classical bits the command/sub-commands read from
        """
        logger.debug(
            "Resolving shard dependencies with qubits=%s bits_written=%s bits_read=%s",
            qubits,
            bits_written,
            bits_read,
        )

        depends_upon: set[int] = set()

        for qubit in qubits:
            if qubit in self._qubit_touched_by:
                logger.debug(
                    "...adding shard dep %s -> qubit overlap",
                    self._qubit_touched_by[qubit],
                )
                depends_upon.add(self._qubit_touched_by[qubit])

        for bit_read in bits_read:
            if bit_read in self._bit_written_by:
                logger.debug(
                    "...adding shard dep %s -> RAW", self._bit_written_by[bit_read]
                )
                depends_upon.add(self._bit_written_by[bit_read])

        for bit_written in bits_written:
            if bit_written in self._bit_written_by:
                logger.debug(
                    "...adding shard dep %s -> WAW", self._bit_written_by[bit_written]
                )
                depends_upon.add(self._bit_written_by[bit_written])
            if bit_written in self._bit_read_by:
                logger.debug(
                    "...adding shard dep %s -> WAR", self._bit_read_by[bit_written]
                )
                depends_upon.add(self._bit_read_by[bit_written])

        return depends_upon

    def _mark_dependencies(
        self,
        shard: Shard,
    ) -> None:
        """Marks (updates) the dependency maps.

        This allows subsequent shard dependency resolution to have the right
        state of everything, updating the shards

        Args:
            shard: Shard to be updated
        """
        for qubit in shard.qubits_used:
            self._qubit_touched_by[qubit] = shard.ID
        for bit in shard.bits_written:
            self._bit_written_by[bit] = shard.ID
        for bit in shard.bits_read:
            self._bit_read_by[bit] = shard.ID
        logger.debug("... dependencies marked")

    def _cleanup_remaining_commands(self) -> None:
        """Cleans up any remaining subcommands.

        This is done by creating a superfluous Barrier command that serves just
        to roll up lingering subcommands.
        """
        remaining_qubits = [k for k, v in self._pending_commands.items() if v]
        logger.debug(
            "Cleaning up remaining subcommands for qubits %s", remaining_qubits
        )
        for qubit in remaining_qubits:
            logger.debug("Adding barrier for subcommands for qubit %s", qubit)
            self._circuit.add_barrier([qubit])
            # Easiest way to get to a command, since there's no constructor. Could
            # create an entire orphan circuit with the matching qubits and the barrier
            # instead, if this has unintended consequences
            barrier_command = self._circuit.get_commands()[-1]
            self._build_shard(barrier_command)

    def _add_pending_sub_command(self, command: Command) -> None:
        """Adds a pending command.

        Adds a pending sub command to the buffer to be flushed when a schedulable
        operation creates a Shard.

        Args:
            command:  tket command (operation, bits, etc)
        """
        qubit_key = command.qubits[0]
        if qubit_key not in self._pending_commands:
            self._pending_commands[qubit_key] = []
        self._pending_commands[qubit_key].append(command)
        logger.debug("Added pending sub-command %s", command)

    @staticmethod
    def should_op_create_shard(op: Op) -> bool:
        """Decide whether to create a shard.

        This includes non-gate operations like measure/reset as well as 2-qubit gates.

        Args:
            op: operation

        Returns:
            `True` if the operation is one that should result in shard creation
        """
        return (
            op.type in SHARD_TRIGGER_OP_TYPES
            or isinstance(op, Conditional)
            or (op.is_gate() and op.n_qubits > 1)
        )

    @staticmethod
    def _is_command_global_phase(command: Command) -> bool:
        """Check if an operation related to global phase.

        Args:
            command: Command to evaluate
        """
        return command.op.type == OpType.Phase or (
            isinstance(command.op, Conditional) and command.op.op.type == OpType.Phase
        )



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from phir.model import PHIRModel
from rich import print  # noqa: A004

from pytket.phir.machine import Machine, MachineTimings
from pytket.phir.phirgen import genphir
from pytket.phir.place_and_route import place_and_route
from pytket.phir.placement import placement_check
from pytket.phir.qtm_machine import QTM_DEFAULT_GATESET, QTM_MACHINES_MAP, QtmMachine
from pytket.phir.sharding.sharder import Sharder
from tests.test_utils import QasmFile, get_qasm_as_circuit

if __name__ == "__main__":
    machine = Machine(3, QTM_DEFAULT_GATESET, {1}, MachineTimings(3.0, 1.0, 2.0, 2.0))
    # force machine options for this test
    # machines normally don't like odd numbers of qubits
    machine.sq_options = {0, 1, 2}

    h11 = QTM_MACHINES_MAP[QtmMachine.H1]

    circuit = get_qasm_as_circuit(QasmFile.classical_hazards)
    sharder = Sharder(circuit)
    shards = sharder.shard()

    output = place_and_route(shards, h11)
    ez_ops_0 = [[0, 2], [1]]
    ez_ops_1 = [[0], [2]]
    state_0 = output[0][0]
    state_1 = output[1][0]
    assert placement_check(ez_ops_0, machine.tq_options, machine.sq_options, state_0)
    assert placement_check(ez_ops_1, machine.tq_options, machine.sq_options, state_1)
    shards_0 = output[0][1]
    shards_1 = output[1][1]
    for shard in shards_0:
        assert shard.ID in {0, 1}
    for shard in shards_1:
        assert shard.ID in {2, 3}
    cost_0 = output[0][2]
    cost_1 = output[1][2]
    assert cost_1 == 0.0

    phir_json = genphir(output, circuit)

    print(PHIRModel.model_validate_json(phir_json))  # type: ignore[misc]



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

import logging

from pytket.circuit import Circuit, OpType
from pytket.phir.qtm_machine import QtmMachine
from pytket.phir.rebasing.rebaser import rebase_to_qtm_machine

from .test_utils import QasmFile, get_qasm_as_circuit

EXPECTED_GATES = [
    OpType.Measure,
    OpType.Rz,
    OpType.PhasedX,
    OpType.ZZPhase,
]


logger = logging.getLogger(__name__)


class TestRebaser:
    def test_rebaser_happy_path_arc1a(self) -> None:
        circ = get_qasm_as_circuit(QasmFile.baby)
        rebased: Circuit = rebase_to_qtm_machine(circ, QtmMachine.H1)

        logger.info(rebased)
        for command in rebased.get_commands():
            if command.op.is_gate():
                assert command.op.type in EXPECTED_GATES



##############################################################################
#
# Copyright (c) 2023-2024 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

import json
from enum import Enum, auto
from pathlib import Path
from typing import TYPE_CHECKING

from wasmtime import wat2wasm

from pytket.phir.phirgen import WORDSIZE
from pytket.phir.phirgen_parallel import genphir_parallel
from pytket.phir.place_and_route import place_and_route
from pytket.phir.qtm_machine import QTM_MACHINES_MAP, QtmMachine
from pytket.phir.rebasing.rebaser import rebase_to_qtm_machine
from pytket.phir.sharding.sharder import Sharder
from pytket.qasm.qasm import circuit_from_qasm

if TYPE_CHECKING:
    from pytket.circuit import Circuit
    from pytket.phir.phirgen import JsonDict


class QasmFile(Enum):
    baby = auto()
    simple = auto()
    eztest = auto()
    baby_with_rollup = auto()
    big_gate = auto()
    simple_cond = auto()
    n10_test = auto()
    classical_hazards = auto()
    cond_1 = auto()
    barrier_complex = auto()
    cond_classical = auto()
    bv_n10 = auto()
    oned_brickwork_circuit_n20 = auto()
    qv20_0 = auto()
    parallelization_test = auto()
    tk2_same_angle = auto()
    tk2_diff_angles = auto()
    rxrz = auto()
    classical_ordering = auto()
    single_qubit_parallel_test = auto()
    exec_order_two_qubits = auto()
    rz_exec_order_three_qubits = auto()
    cond_barrier = auto()
    arbitrary_qreg_names = auto()
    group_ordering = auto()
    sleep = auto()
    classical0 = auto()
    classical1 = auto()


class WatFile(Enum):
    add = auto()
    testfile = auto()


def get_qasm_as_circuit(
    qasm_file: QasmFile,
    *,
    use_clexpr: bool = False,
) -> "Circuit":
    """Utility function to convert a QASM file to Circuit.

    Args:
        qasm_file: enum for a QASM file
        use_clexpr: convert classical expressions to ClExprOp operations

    Returns:
        Corresponding tket circuit
    """
    this_dir = Path(Path(__file__).resolve()).parent
    return circuit_from_qasm(
        f"{this_dir}/data/qasm/{qasm_file.name}.qasm",
        maxwidth=WORDSIZE,
        use_clexpr=use_clexpr,
    )


def get_phir_json(qasmfile: QasmFile, *, rebase: bool) -> "JsonDict":
    """Get the QASM file for the specified circuit."""
    qtm_machine = QtmMachine.H1
    circuit = get_qasm_as_circuit(qasmfile)
    if rebase:
        circuit = rebase_to_qtm_machine(circuit, qtm_machine)
    machine = QTM_MACHINES_MAP.get(qtm_machine)
    assert machine
    shards = Sharder(circuit).shard()
    placed = place_and_route(shards, machine)
    return json.loads(genphir_parallel(placed, circuit, machine))  # type: ignore[misc, no-any-return]


def get_wat_as_wasm_bytes(wat_file: WatFile) -> bytes:
    """Gets a given wat file, converted to WASM bytes by wasmtime."""
    this_dir = Path(Path(__file__).resolve()).parent
    return wat2wasm(
        Path(f"{this_dir}/data/wasm/{wat_file.name}.wat").read_text(encoding="utf-8")
    )



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from enum import Enum

from pytket.circuit import OpType

from .machine import Machine, MachineTimings


class QtmMachine(Enum):
    """Available machine architectures."""

    H1 = "H1"


QTM_DEFAULT_GATESET = {OpType.Rz, OpType.PhasedX, OpType.ZZPhase}

QTM_MACHINES_MAP = {
    QtmMachine.H1: Machine(
        size=20,
        gateset=QTM_DEFAULT_GATESET,
        tq_options={0, 2, 4, 6, 8, 10, 12, 14, 16, 18},
        timings=MachineTimings(
            tq_time=0.04,
            sq_time=0.03,
            qb_swap_time=0.9,
            meas_prep_time=0.05,
        ),
    )
}



##############################################################################
#
# Copyright (c) 2023-2024 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"

import logging
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

from rich import print  # noqa: A004

from phir.model import PHIRModel
from pytket.qasm.qasm import circuit_from_qasm_str, circuit_from_qasm_wasm

from .phirgen import WORDSIZE, genphir
from .phirgen_parallel import genphir_parallel
from .place_and_route import place_and_route
from .qtm_machine import QTM_MACHINES_MAP, QtmMachine
from .rebasing.rebaser import rebase_to_qtm_machine
from .sharding.sharder import Sharder

if TYPE_CHECKING:
    from pytket.circuit import Circuit

    from .machine import Machine

logger = logging.getLogger(__name__)


def pytket_to_phir(circuit: "Circuit", qtm_machine: QtmMachine | None = None) -> str:
    """Converts a pytket circuit into its PHIR representation.

    This can optionally include rebasing against a Quantinuum machine architecture,
    and control of the TKET optimization level.

    :param circuit: Circuit object to be converted
    :param qtm_machine: (Optional) Quantinuum machine architecture to rebase against

    Returns:
        PHIR JSON as a str
    """
    logger.info("Starting phir conversion process for circuit %s", circuit)
    machine: Machine | None = None
    if qtm_machine:
        logger.info("Rebasing to machine %s", qtm_machine)
        circuit = rebase_to_qtm_machine(circuit, qtm_machine)
        machine = QTM_MACHINES_MAP.get(qtm_machine)
    else:
        machine = None

    logger.debug("Sharding input circuit...")
    shards = Sharder(circuit).shard()

    if machine:
        # Only print message if a machine object is passed
        # Otherwise, placement and routing are functionally skipped
        # The function is called, but the output is just filled with 0s
        logger.debug("Performing placement and routing...")
    placed = place_and_route(shards, machine)
    # safety check: never run with parallelization on a 1 qubit circuit
    if machine and len(circuit.qubits) > 1:
        phir_json = genphir_parallel(placed, circuit, machine)
    else:
        phir_json = genphir(placed, circuit, machine_ops=bool(machine))
    if logger.getEffectiveLevel() <= logging.INFO:
        print("PHIR JSON:")
        print(PHIRModel.model_validate_json(phir_json))
    return phir_json


def qasm_to_phir(
    qasm: str,
    qtm_machine: QtmMachine | None = None,
    wasm_bytes: bytes | None = None,
) -> str:
    """Converts a QASM circuit string into its PHIR representation.

    This can optionally include rebasing against a Quantinuum machine architecture,
    and control of the TKET optimization level.

    :param qasm: QASM input to be converted
    :param qtm_machine: (Optional) Quantinuum machine architecture to rebase against
    :param wasm_bytes: (Optional) WASM as bytes to include as part of circuit
    """
    circuit: Circuit
    if wasm_bytes:
        with (
            NamedTemporaryFile(suffix=".qasm", delete=False) as qasm_file,
            NamedTemporaryFile(suffix=".wasm", delete=False) as wasm_file,
        ):
            qasm_file.write(qasm.encode())
            qasm_file.flush()
            wasm_file.write(wasm_bytes)
            wasm_file.flush()

            circuit = circuit_from_qasm_wasm(
                qasm_file.name, wasm_file.name, maxwidth=WORDSIZE
            )
    else:
        circuit = circuit_from_qasm_str(qasm, maxwidth=WORDSIZE)
    return pytket_to_phir(circuit, qtm_machine)



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from __future__ import annotations


class TransportError(Exception):
    """Error raised by inverse() util function."""

    def __init__(self, a: list[int], b: list[int]):
        super().__init__(f"Traps different sizes: {len(a)} vs. {len(b)}")


class PermutationError(Exception):
    """Error raised by inverse() util function."""

    def __init__(self, lst: list[int]):
        super().__init__(f"List {lst} is not a permutation of range({len(lst)})")


def inverse(lst: list[int]) -> list[int]:
    """Inverse of a permutation list.

    If a[i] = x, then inverse(a)[x] = i.
    """
    inv = [-1] * len(lst)

    for i, elem in enumerate(lst):
        if not 0 <= elem < len(lst) or inv[elem] != -1:
            raise PermutationError(lst)
        inv[elem] = i

    return inv


def transport_cost(init: list[int], goal: list[int], swap_cost: float) -> float:
    """Cost of transport from init to goal.

    This is based on the number of parallel swaps performed by Odd-Even
    Transposition Sort, which is the maximum distance that any qubit travels.
    """
    if len(init) != len(goal):
        raise TransportError(init, goal)

    n_swaps = max(abs(g - i) for i, g in zip(inverse(init), inverse(goal), strict=True))

    return n_swaps * swap_cost



##############################################################################
#
# Copyright (c) 2024 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"

import json

import pytest

from pytket.circuit import Bit, Circuit
from pytket.circuit.logic_exp import BitWiseOp, create_bit_logic_exp
from pytket.phir.api import pytket_to_phir
from pytket.phir.phirgen import WORDSIZE
from pytket.qasm.qasm import circuit_from_qasm_str
from pytket.unit_id import BitRegister

from .test_utils import QasmFile, get_qasm_as_circuit


def test_multiple_sleep() -> None:
    """Ensure multiple sleep ops get converted correctly."""
    qasm = """
    OPENQASM 2.0;
    include "hqslib1_dev.inc";

    qreg q[2];

    sleep(1) q[0];
    sleep(2) q[1];
    """
    circ = circuit_from_qasm_str(qasm, maxwidth=WORDSIZE)
    phir = json.loads(pytket_to_phir(circ))
    assert phir["ops"][2] == {"mop": "Idle", "args": [["q", 0]], "duration": [1.0, "s"]}
    assert phir["ops"][4] == {"mop": "Idle", "args": [["q", 1]], "duration": [2.0, "s"]}


def test_simple_cond_classical() -> None:
    """Ensure conditional classical operation are correctly generated."""
    circ = get_qasm_as_circuit(QasmFile.simple_cond)
    phir = json.loads(pytket_to_phir(circ))
    assert phir["ops"][-6] == {"//": "IF ([c[0]] == 1) THEN SetBits(1) z[0];"}
    assert phir["ops"][-5] == {
        "block": "if",
        "condition": {"cop": "==", "args": [["c", 0], 1]},
        "true_branch": [{"cop": "=", "returns": [["z", 0]], "args": [1]}],
    }


def test_pytket_classical_only() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/61 ."""
    c = Circuit()
    a = c.add_c_register("a", 2)
    b = c.add_c_register("b", 3)

    c.add_c_copyreg(a, b)
    c.add_c_copybits([Bit("b", 2), Bit("a", 1)], [Bit("a", 0), Bit("b", 0)])
    c.add_c_copybits(
        [Bit("b", 2), Bit("a", 1)], [Bit("a", 0), Bit("b", 0)], condition=Bit("b", 1)
    )
    c.add_c_copybits(
        [Bit("a", 0), Bit("a", 1)],  # type: ignore[list-item] # overloaded function
        [Bit("b", 0), Bit("b", 1)],  # type: ignore[list-item] # overloaded function
        condition_bits=[Bit("b", 1), Bit("b", 2)],
        condition_value=2,
    )

    phir = json.loads(pytket_to_phir(c))

    assert phir["ops"][3] == {
        "cop": "=",
        "returns": [["b", 0], ["b", 1]],
        "args": [["a", 0], ["a", 1]],
    }
    assert phir["ops"][5] == {
        "cop": "=",
        "returns": [["a", 0], ["b", 0]],
        "args": [["b", 2], ["a", 1]],
    }
    assert phir["ops"][7] == {
        "block": "if",
        "condition": {"cop": "==", "args": [["b", 1], 1]},
        "true_branch": [
            {"cop": "=", "returns": [["a", 0], ["b", 0]], "args": [["b", 2], ["a", 1]]}
        ],
    }
    assert phir["ops"][8] == {
        "//": "IF ([b[1], b[2]] == 2) THEN CopyBits a[0], a[1], b[0], b[1];"
    }
    assert phir["ops"][9] == {
        "block": "if",
        "condition": {
            "cop": "&",
            "args": [
                {"cop": "==", "args": [["b", 1], 0]},
                {"cop": "==", "args": [["b", 2], 1]},
            ],
        },
        "true_branch": [
            {"cop": "=", "returns": [["b", 0], ["b", 1]], "args": [["a", 0], ["a", 1]]}
        ],
    }


def test_classicalexpbox() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/86 ."""
    circ = Circuit()
    a = circ.add_c_register("a", 2)
    b = circ.add_c_register("b", 2)
    c = circ.add_c_register("c", 3)
    circ.add_classicalexpbox_register(a + b, c.to_list())

    phir = json.loads(pytket_to_phir(circ))
    assert phir["ops"][4] == {
        "cop": "=",
        "returns": ["c"],
        "args": [{"cop": "+", "args": ["a", "b"]}],
    }


def test_nested_arith() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/87 ."""
    circ = Circuit()
    a = circ.add_c_register("a", 2)
    b = circ.add_c_register("b", 2)
    c = circ.add_c_register("c", 3)
    circ.add_classicalexpbox_register(a + b // c, c.to_list())

    phir = json.loads(pytket_to_phir(circ))
    assert phir["ops"][4] == {
        "cop": "=",
        "returns": ["c"],
        "args": [{"cop": "+", "args": ["a", {"cop": "/", "args": ["b", "c"]}]}],
    }


def test_arith_with_int() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/88 ."""
    circ = Circuit()
    a = circ.add_c_register("a", 2)
    circ.add_classicalexpbox_register(a << 1, a.to_list())

    phir = json.loads(pytket_to_phir(circ))
    assert phir["ops"][2] == {
        "cop": "=",
        "returns": ["a"],
        "args": [{"cop": "<<", "args": ["a", 1]}],
    }


def test_bitwise_ops() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/91 ."""
    circ = Circuit()
    a = circ.add_c_register("a", 2)
    b = circ.add_c_register("b", 2)
    c = circ.add_c_register("c", 1)
    expr = a[0] ^ b[0]
    circ.add_classicalexpbox_bit(expr, [c[0]])

    phir = json.loads(pytket_to_phir(circ))
    assert phir["ops"][4] == {
        "cop": "=",
        "returns": [["c", 0]],
        "args": [{"cop": "^", "args": [["a", 0], ["b", 0]]}],
    }


def test_conditional_barrier() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/119 ."""
    circ = get_qasm_as_circuit(QasmFile.cond_barrier)
    phir = json.loads(pytket_to_phir(circ))
    assert phir["ops"][4] == {"//": "IF ([m[0], m[1]] == 0) THEN Barrier q[0], q[1];"}
    assert phir["ops"][5] == {
        "block": "if",
        "condition": {
            "cop": "&",
            "args": [
                {"cop": "==", "args": [["m", 0], 0]},
                {"cop": "==", "args": [["m", 1], 0]},
            ],
        },
        "true_branch": [{"meta": "barrier", "args": [["q", 0], ["q", 1]]}],
    }


def test_nested_bitwise_op() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/133 ."""
    circ = Circuit()
    a = circ.add_c_register("a", 4)
    b = circ.add_c_register("b", 1)
    circ.add_classicalexpbox_bit(a[0] ^ a[1] ^ a[2] ^ a[3], [b[0]])

    phir = json.loads(pytket_to_phir(circ))
    assert phir["ops"][3] == {
        "cop": "=",
        "returns": [["b", 0]],
        "args": [
            {
                "cop": "^",
                "args": [
                    {
                        "cop": "^",
                        "args": [{"cop": "^", "args": [["a", 0], ["a", 1]]}, ["a", 2]],
                    },
                    ["a", 3],
                ],
            }
        ],
    }


def test_sleep_idle() -> None:
    """Ensure sleep from qasm gets converted to PHIR Idle Mop."""
    circ = get_qasm_as_circuit(QasmFile.sleep)
    phir = json.loads(pytket_to_phir(circ))
    assert phir["ops"][7] == {"mop": "Idle", "args": [["q", 0]], "duration": [1.0, "s"]}


def test_reordering_classical_conditional() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/150 ."""
    circuit = Circuit(1)

    ctrl = circuit.add_c_register(name="ctrl", size=1)
    meas = circuit.add_c_register(name="meas", size=1)

    circuit.add_c_setreg(1, ctrl)
    circuit.X(0, condition=ctrl[0])

    circuit.add_c_setreg(0, ctrl)
    circuit.X(0, condition=ctrl[0])

    circuit.Measure(
        qubit=circuit.qubits[0],
        bit=meas[0],
    )

    phir = json.loads(pytket_to_phir(circuit))

    assert phir["ops"][4] == {"args": [1], "cop": "=", "returns": [["ctrl", 0]]}
    assert phir["ops"][6] == {
        "block": "if",
        "condition": {"args": [["ctrl", 0], 1], "cop": "=="},
        "true_branch": [{"angles": None, "args": [["q", 0]], "qop": "X"}],
    }
    assert phir["ops"][8] == {"args": [0], "cop": "=", "returns": [["ctrl", 0]]}
    assert phir["ops"][10] == {
        "block": "if",
        "condition": {"args": [["ctrl", 0], 1], "cop": "=="},
        "true_branch": [{"angles": None, "args": [["q", 0]], "qop": "X"}],
    }


def test_conditional_measure() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/154 ."""
    c = Circuit(2, 2)
    c.H(0).H(1)
    c.Measure(0, 0)
    c.Measure(1, 1, condition_bits=[0], condition_value=1)
    phir = json.loads(pytket_to_phir(c))
    assert phir["ops"][-2] == {"//": "IF ([c[0]] == 1) THEN Measure q[1] --> c[1];"}
    assert phir["ops"][-1] == {
        "block": "if",
        "condition": {"cop": "==", "args": [["c", 0], 1]},
        "true_branch": [{"qop": "Measure", "returns": [["c", 1]], "args": [["q", 1]]}],
    }


def test_conditional_classical_not() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/159 ."""
    circuit = Circuit()
    target_reg = circuit.add_c_register(BitRegister(name="target_reg", size=1))
    control_reg = circuit.add_c_register(BitRegister(name="control_reg", size=1))

    circuit.add_c_not(
        arg_in=target_reg[0], arg_out=target_reg[0], condition=control_reg[0]
    )

    phir = json.loads(pytket_to_phir(circuit))
    assert phir["ops"][-1] == {
        "block": "if",
        "condition": {"cop": "==", "args": [["control_reg", 0], 1]},
        "true_branch": [
            {
                "cop": "=",
                "returns": [["target_reg", 0]],
                "args": [{"cop": "~", "args": [["target_reg", 0]]}],
            }
        ],
    }


def test_explicit_classical_ops() -> None:
    """Test explicit predicates and modifiers."""
    # From https://github.com/CQCL/tket/blob/a2f6fab8a57da8787dfae94764b7c3a8e5779024/pytket/tests/classical_test.py#L97-L101
    c = Circuit(0, 4)
    # predicates
    c.add_c_and(1, 2, 3)
    c.add_c_not(0, 1)
    c.add_c_xor(1, 2, 3)
    # modifiers
    c.add_c_and(2, 3, 3)
    c.add_c_or(0, 3, 0)
    phir = json.loads(pytket_to_phir(c))
    assert phir["ops"][1] == {"//": "AND c[1], c[2], c[3];"}
    assert phir["ops"][2] == {
        "cop": "=",
        "returns": [["c", 3]],
        "args": [{"cop": "&", "args": [["c", 1], ["c", 2]]}],
    }
    assert phir["ops"][3] == {"//": "NOT c[0], c[1];"}
    assert phir["ops"][4] == {
        "cop": "=",
        "returns": [["c", 1]],
        "args": [{"cop": "~", "args": [["c", 0]]}],
    }
    assert phir["ops"][5] == {"//": "XOR c[1], c[2], c[3];"}
    assert phir["ops"][6] == {
        "cop": "=",
        "returns": [["c", 3]],
        "args": [{"cop": "^", "args": [["c", 1], ["c", 2]]}],
    }
    assert phir["ops"][7] == {"//": "AND c[2], c[3];"}
    assert phir["ops"][8] == {
        "cop": "=",
        "returns": [["c", 3]],
        "args": [{"cop": "&", "args": [["c", 2], ["c", 3]]}],
    }
    assert phir["ops"][9] == {"//": "OR c[3], c[0];"}
    assert phir["ops"][10] == {
        "cop": "=",
        "returns": [["c", 0]],
        "args": [{"cop": "|", "args": [["c", 3], ["c", 0]]}],
    }


def test_multi_bit_ops() -> None:
    """Test classical ops added to the circuit via tket multi-bit ops."""
    # Test from https://github.com/CQCL/tket/blob/a2f6fab8a57da8787dfae94764b7c3a8e5779024/pytket/tests/classical_test.py#L107-L112
    c = Circuit()
    c0 = c.add_c_register("c0", 3)
    c1 = c.add_c_register("c1", 4)
    c2 = c.add_c_register("c2", 5)
    # predicates
    c.add_c_and_to_registers(c0, c1, c2)
    c.add_c_not_to_registers(c1, c2)
    c.add_c_or_to_registers(c0, c1, c2)
    # modifier
    c.add_c_xor_to_registers(c2, c1, c2)
    # conditionals
    c.add_c_not_to_registers(c1, c2, condition=Bit("c0", 0))
    c.add_c_not_to_registers(c1, c1, condition=Bit("c0", 0))
    phir = json.loads(pytket_to_phir(c))
    assert phir["ops"][3] == {
        "//": "AND (*3) c0[0], c1[0], c2[0], c0[1], c1[1], c2[1], c0[2], c1[2], c2[2];"
    }
    assert phir["ops"][4] == {
        "cop": "=",
        "returns": ["c2"],
        "args": [{"cop": "&", "args": ["c0", "c1"]}],
    }
    assert phir["ops"][5] == {
        "//": "NOT (*4) c1[0], c2[0], c1[1], c2[1], c1[2], c2[2], c1[3], c2[3];"
    }
    assert phir["ops"][6] == {
        "cop": "=",
        "returns": ["c2"],
        "args": [{"cop": "~", "args": ["c1"]}],
    }
    assert phir["ops"][7] == {
        "//": "OR (*3) c0[0], c1[0], c2[0], c0[1], c1[1], c2[1], c0[2], c1[2], c2[2];"
    }
    assert phir["ops"][8] == {
        "cop": "=",
        "returns": ["c2"],
        "args": [{"cop": "|", "args": ["c0", "c1"]}],
    }
    assert phir["ops"][9] == {
        "//": "XOR (*4) c1[0], c2[0], c1[1], c2[1], c1[2], c2[2], c1[3], c2[3];"
    }
    assert phir["ops"][10] == {
        "cop": "=",
        "returns": ["c2"],
        "args": [{"cop": "^", "args": ["c1", "c2"]}],
    }
    assert phir["ops"][12] == {
        "block": "if",
        "condition": {"cop": "==", "args": [["c0", 0], 1]},
        "true_branch": [
            {"cop": "=", "returns": ["c2"], "args": [{"cop": "~", "args": ["c1"]}]}
        ],
    }
    assert phir["ops"][14] == {
        "block": "if",
        "condition": {"cop": "==", "args": [["c0", 0], 1]},
        "true_branch": [
            {"cop": "=", "returns": ["c1"], "args": [{"cop": "~", "args": ["c1"]}]}
        ],
    }


def test_irregular_multibit_ops() -> None:
    """From https://github.com/CQCL/pytket-phir/pull/162#discussion_r1555807863 ."""
    c = Circuit()
    areg = c.add_c_register("a", 2)
    breg = c.add_c_register("b", 2)
    creg = c.add_c_register("c", 2)
    c.add_c_and_to_registers(areg, breg, creg)
    mbop = c.get_commands()[0].op
    c.add_gate(mbop, [areg[0], areg[1], breg[0], breg[1], creg[0], creg[1]])

    phir = json.loads(pytket_to_phir(c))
    assert phir["ops"][3] == {"//": "AND (*2) a[0], b[0], c[0], a[1], b[1], c[1];"}
    assert phir["ops"][4] == {
        "cop": "=",
        "returns": ["c"],
        "args": [{"cop": "&", "args": ["a", "b"]}],
    }
    assert phir["ops"][5] == {"//": "AND (*2) a[0], a[1], b[0], b[1], c[0], c[1];"}
    assert phir["ops"][6] == {
        "block": "sequence",
        "ops": [
            {
                "cop": "=",
                "returns": [["b", 0]],
                "args": [{"cop": "&", "args": [["a", 0], ["a", 1]]}],
            },
            {
                "cop": "=",
                "returns": [["c", 1]],
                "args": [{"cop": "&", "args": [["b", 1], ["c", 0]]}],
            },
        ],
    }


def test_nullary_ops() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/178 ."""
    c = Circuit(1, 1)
    exp1 = create_bit_logic_exp(BitWiseOp.ONE, [])
    c.H(0, condition=exp1)
    exp0 = create_bit_logic_exp(BitWiseOp.ZERO, [])
    c.H(0, condition=exp0)
    c.measure_all()
    phir = json.loads(pytket_to_phir(c))

    assert phir["ops"][4] == {
        "cop": "=",
        "returns": [["tk_SCRATCH_BIT", 0]],
        "args": [1],
    }
    assert phir["ops"][6] == {
        "cop": "=",
        "returns": [["tk_SCRATCH_BIT", 1]],
        "args": [0],
    }
    assert phir["ops"][8]["condition"] == {
        "cop": "==",
        "args": [["tk_SCRATCH_BIT", 0], 1],
    }
    assert phir["ops"][10]["condition"] == {
        "cop": "==",
        "args": [["tk_SCRATCH_BIT", 1], 1],  # evals to False
    }


def test_condition_multiple_bits() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/215 ."""
    n_bits = 3
    c = Circuit(1, n_bits)
    c.Rz(0.5, 0, condition_bits=list(range(n_bits)), condition_value=6)
    phir = json.loads(pytket_to_phir(c))

    assert phir["ops"][2] == {"//": "IF ([c[0], c[1], c[2]] == 6) THEN Rz(0.5) q[0];"}
    assert phir["ops"][3]["condition"] == {
        "cop": "&",
        "args": [
            {"cop": "==", "args": [["c", 0], 0]},
            {
                "cop": "&",
                "args": [
                    {"cop": "==", "args": [["c", 1], 1]},
                    {"cop": "==", "args": [["c", 2], 1]},
                ],
            },
        ],
    }


def test_unused_classical_registers() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/237 ."""
    circ = Circuit()
    _ = circ.add_c_register("a", 1)
    phir = json.loads(pytket_to_phir(circ))

    assert phir["ops"][0] == {
        "data": "cvar_define",
        "data_type": "i64",
        "size": 1,
        "variable": "a",
    }


@pytest.mark.parametrize("use_clexpr", [False, True])
def test_classical_0(*, use_clexpr: bool) -> None:
    """Test handling of ClassicalExpBox/ClExprOp."""
    circ = get_qasm_as_circuit(QasmFile.classical0, use_clexpr=use_clexpr)
    phir = json.loads(pytket_to_phir(circ))
    ops = phir["ops"]
    assert {
        "cop": "=",
        "returns": ["d"],
        "args": [{"cop": "<<", "args": ["a", 1]}],
    } in ops
    assert {
        "block": "if",
        "condition": {"cop": "==", "args": [["tk_SCRATCH_BIT", 0], 0]},
        "true_branch": [
            {
                "cop": "=",
                "returns": [["c", 1]],
                "args": [
                    {
                        "cop": "|",
                        "args": [
                            {"cop": "&", "args": [["b", 1], ["a", 1]]},
                            ["a", 0],
                        ],
                    }
                ],
            }
        ],
    } in ops
    assert {
        "cop": "=",
        "returns": ["c"],
        "args": [{"cop": "&", "args": ["b", "a"]}],
    } in ops
    assert {
        "cop": "=",
        "returns": ["b"],
        "args": [{"cop": "+", "args": ["a", "b"]}],
    } in ops
    assert {
        "cop": "=",
        "returns": [["b", 1]],
        "args": [{"cop": "^", "args": [["b", 0], {"cop": "~", "args": [["b", 2]]}]}],
    } in ops
    assert {
        "cop": "=",
        "returns": ["c"],
        "args": [{"cop": "-", "args": ["a", {"cop": "*", "args": ["b", "c"]}]}],
    } in ops
    assert {
        "cop": "=",
        "returns": ["d"],
        "args": [{"cop": "<<", "args": ["a", 1]}],
    } in ops
    assert {
        "cop": "=",
        "returns": ["d"],
        "args": [{"cop": ">>", "args": ["c", 2]}],
    } in ops
    assert {
        "cop": "=",
        "returns": ["b"],
        "args": [{"cop": "*", "args": [{"cop": "*", "args": ["a", "c"]}, "b"]}],
    } in ops
    assert {
        "cop": "=",
        "returns": [["d", 0]],
        "args": [{"cop": "^", "args": [["a", 0], 1]}],
    } in ops


@pytest.mark.parametrize("use_clexpr", [False, True])
def test_classical_1(*, use_clexpr: bool) -> None:
    """Test handling of ClassicalExpBox/ClExprOp."""
    circ = get_qasm_as_circuit(QasmFile.classical1, use_clexpr=use_clexpr)
    phir = json.loads(pytket_to_phir(circ))
    ops = phir["ops"]
    assert {
        "block": "if",
        "condition": {"cop": "==", "args": [["tk_SCRATCH_BIT", 0], 0]},
        "true_branch": [
            {
                "cop": "=",
                "returns": [["c", 1]],
                "args": [
                    {
                        "cop": "|",
                        "args": [
                            {"cop": "&", "args": [["b", 1], ["a", 1]]},
                            ["a", 0],
                        ],
                    }
                ],
            }
        ],
    } in ops
    assert {
        "cop": "=",
        "returns": ["c"],
        "args": [{"cop": "|", "args": [{"cop": "&", "args": ["b", "a"]}, "d"]}],
    } in ops
    assert {
        "cop": "=",
        "returns": [["d", 0]],
        "args": [{"cop": "^", "args": [["a", 0], 1]}],
    } in ops



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"

import logging

import pytest

from pytket.phir.api import pytket_to_phir, qasm_to_phir
from pytket.phir.qtm_machine import QtmMachine

from .test_utils import QasmFile, get_qasm_as_circuit

logger = logging.getLogger(__name__)


class TestApi:
    def test_pytket_to_phir_no_machine(self) -> None:
        """Test case when no machine is present."""
        circuit = get_qasm_as_circuit(QasmFile.baby)
        assert pytket_to_phir(circuit)

    @pytest.mark.parametrize("test_file", list(QasmFile))
    def test_pytket_to_phir_no_machine_all(self, test_file: QasmFile) -> None:
        """Test case when no machine is present."""
        circuit = get_qasm_as_circuit(test_file)

        match test_file:
            case QasmFile.big_gate:
                with pytest.raises(KeyError, match=r".*CnX.*"):
                    assert pytket_to_phir(circuit)
            case QasmFile.qv20_0:
                with pytest.raises(KeyError, match=r".*U3.*"):
                    assert pytket_to_phir(circuit)
            case _:
                assert pytket_to_phir(circuit)

    @pytest.mark.parametrize("test_file", list(QasmFile))
    def test_pytket_to_phir_h1_1_all(self, test_file: QasmFile) -> None:
        """Standard case."""
        circuit = get_qasm_as_circuit(test_file)

        assert pytket_to_phir(circuit, QtmMachine.H1)

    def test_qasm_to_phir(self) -> None:
        """Test the qasm string entrypoint works."""
        qasm = """
        OPENQASM 2.0;
        include "qelib1.inc";

        qreg q[3];
        h q;
        ZZ q[1],q[0];
        creg cr[3];
        measure q[0]->cr[0];
        measure q[1]->cr[0];
        """

        assert qasm_to_phir(qasm, QtmMachine.H1)



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

import io
from dataclasses import dataclass, field
from itertools import count
from typing import TYPE_CHECKING, TypeAlias

if TYPE_CHECKING:
    from pytket.circuit import Command
    from pytket.unit_id import Bit, Qubit, UnitID


@dataclass(frozen=True)
class Shard:
    """The Shard class.

    A shard is a logical grouping of operations that represents the unit by which
    we actually do placement of qubits.
    """

    # The unique identifier of the shard
    ID: int = field(default_factory=count().__next__, init=False)

    # The "schedulable" command of the shard
    primary_command: "Command"

    # The other commands related to the primary schedulable command, stored
    # as a map of bit-handle (unitID) -> list[Command]
    sub_commands: dict["UnitID", list["Command"]]

    # All qubits used by the primary and sub commands
    qubits_used: set["Qubit"]

    # Set of all classical bits written to by the primary and sub commands
    bits_written: set["Bit"]

    # Set of all classical bits read by the primary and sub commands
    bits_read: set["Bit"]

    # A set of the identifiers of other shards this particular shard depends upon
    depends_upon: set[int]

    def __hash__(self) -> int:
        """Hashing for shards is done only by its autogen unique int ID."""
        return self.ID

    def pretty_print(self) -> str:
        """Returns the shard in a human-friendly format."""
        output = io.StringIO()
        output.write(f"Shard {self.ID}:")
        output.write(f"\n   Command: {self.primary_command}")
        output.write("\n   Sub commands: ")
        if not self.sub_commands:
            output.write("[]")
        for sub in self.sub_commands:
            output.write(f"\n       {sub}: {self.sub_commands[sub]}")
        output.write(f"\n   Qubits used:  {self.qubits_used}")
        output.write(f"\n   Bits written: {self.bits_written}")
        output.write(f"\n   Bits read:    {self.bits_read}")
        output.write(f"\n   Depends upon: {self.depends_upon}")
        content = output.getvalue()
        output.close()
        return content


Cost: TypeAlias = float
ShardLayer: TypeAlias = list[Shard]
Ordering: TypeAlias = list[int]



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

import bisect
import math

from .routing import inverse


class GateOpportunitiesError(Exception):
    """Exception raised when gating zones cannot accommodate all operations."""

    def __init__(self) -> None:
        super().__init__("Not enough gating opportunities for all ops in this layer")


class InvalidParallelOpsError(Exception):
    """Raised when a layer tries to gate the same qubit more than once in parallel."""

    def __init__(self, q: int) -> None:
        super().__init__(f"Cannot gate qubit {q} more than once in the same layer")


class PlacementCheckError(Exception):
    """Exception raised when placement check fails."""

    def __init__(self) -> None:
        super().__init__("Placement Check Failed")


def placement_check(
    ops: list[list[int]],
    tq_options: set[int],
    sq_options: set[int],
    state: list[int],
) -> bool:
    """Ensure that the qubits end up in the right gating zones."""
    placement_valid = False
    inv = inverse(state)

    # If there are no operations to place, it does not matter where the
    # qubits are and any placement is valid
    if not ops:
        return True

    # assume ops look like this [[1,2],[3],[4],[5,6],[7],[8],[9,10]]
    for op in ops:
        if len(op) == 2:  # tq operation   # noqa: PLR2004
            q1, q2 = op[0], op[1]
            # check that the q1 is next to q2 and they are in the right zone
            zone = (inv[q1] in tq_options) | (inv[q2] in tq_options)
            neighbor = (state.index(q2) == state.index(q1) + 1) | (
                state.index(q1) == state.index(q2) + 1
            )
            placement_valid = zone & neighbor

        else:  # sq operation
            q = op[0]
            zone = state.index(q) in sq_options
            placement_valid = zone

    return placement_valid


def nearest(zone: int, options: set[int]) -> int:
    """Return the nearest available zone to the given zone."""
    lst = sorted(options)
    ind = bisect.bisect_left(lst, zone)

    if ind == 0:
        nearest_zone = lst[0]
    elif ind == len(lst):
        nearest_zone = lst[-1]
    else:
        lft = lst[ind - 1]
        rgt = lst[ind]
        nearest_zone = lft if rgt - zone > zone - lft else rgt

    return nearest_zone


def place_tq_ops(
    tq_ops: list[list[int]],
    placed_qubits: set[int],
    order: list[int],
    tq_zones: set[int],
    sq_zones: set[int],
) -> list[int]:
    """A helper function to place the TQ operations."""
    for op in tq_ops:
        q1, q2 = op[0], op[1]
        # check to make sure that the qubits have not already been placed
        if q1 in placed_qubits:
            raise InvalidParallelOpsError(q1)
        if q2 in placed_qubits:
            raise InvalidParallelOpsError(q2)
        midpoint = math.floor(abs(q2 - q1) / 2) + min(q1, q2)
        # find the tq gating zone closest to the midpoint of the 2 qubits
        nearest_tq_zone = nearest(midpoint, tq_zones)
        order[nearest_tq_zone] = q1
        order[nearest_tq_zone + 1] = q2
        # remove the occupied zones in the tap from tq and sq options
        tq_zones.discard(nearest_tq_zone)
        tq_zones.discard(nearest_tq_zone + 1)
        sq_zones.discard(nearest_tq_zone)
        sq_zones.discard(nearest_tq_zone + 1)
        placed_qubits.add(q1)
        placed_qubits.add(q2)
    return order


def place(  # noqa: PLR0912
    ops: list[list[int]],
    tq_options: set[int],
    sq_options: set[int],
    num_qubits: int,
) -> list[int]:
    """Place the qubits in the right order."""
    # assume ops look like this [[1,2],[3],[4],[5,6],[7],[8],[9,10]]
    order = [-1] * num_qubits
    placed_qubits: set[int] = set()

    tq_zones = tq_options.copy()
    sq_zones = sq_options.copy()

    tq_ops = []
    sq_ops = []

    # get separate lists of tq and sq operations
    for op in ops:
        if len(op) == 2:  # tq operation  # noqa: PLR2004
            tq_ops.append(op)
        else:  # sq_operation
            sq_ops.append(op)

    # sort the tq_ops by distance apart [[furthest] -> [closest]]
    tq_ops_sorted = sorted(tq_ops, key=lambda x: abs(x[0] - x[1]), reverse=True)  # type: ignore [misc]

    # check to make sure that there are zones available for all ops
    if len(tq_ops) > len(tq_zones):
        raise GateOpportunitiesError
    if len(sq_ops) > len(sq_zones) - 2 * len(tq_ops):
        # Because SQ zones are offsets of TQ zones, each tq op covers 2 sq zones
        raise GateOpportunitiesError

    # place the tq ops
    order = place_tq_ops(tq_ops_sorted, placed_qubits, order, tq_zones, sq_zones)

    # place the sq ops
    for op in sq_ops:
        q1 = op[0]
        # check to make sure that the qubits have not already been placed
        if q1 in placed_qubits:
            raise InvalidParallelOpsError(q1)
        # place the qubit in the first available zone
        for i in range(num_qubits):
            if i in sq_zones:
                order[i] = q1
                tq_zones.discard(i)
                sq_zones.discard(i)
                placed_qubits.add(q1)
                break

    # fill in the rest of the slots in the order with the inactive qubits
    for i in range(num_qubits):
        if i not in placed_qubits:
            for j in range(num_qubits):
                if order[j] == -1:
                    order[j] = i
                    break

    if placement_check(ops, tq_options, sq_options, order):
        return order

    raise PlacementCheckError


def optimized_place(
    ops: list[list[int]],
    tq_options: set[int],
    sq_options: set[int],
    num_qubits: int,
    prev_state: list[int],
) -> list[int]:
    """Place the qubits in the right order."""
    # assume ops look like this [[1,2],[3],[4],[5,6],[7],[8],[9,10]]
    order = [-1] * num_qubits
    placed_qubits: set[int] = set()

    tq_zones = tq_options.copy()
    sq_zones = sq_options.copy()

    tq_ops = []
    sq_ops = []

    # get separate lists of tq and sq operations
    for op in ops:
        if len(op) == 2:  # tq operation  # noqa: PLR2004
            tq_ops.append(op)
        else:  # sq_operation
            sq_ops.append(op)

    # sort the tq_ops by distance apart [[furthest] -> [closest]]
    tq_ops_sorted = sorted(tq_ops, key=lambda x: abs(x[0] - x[1]), reverse=True)  # type: ignore [misc]

    # check to make sure that there are zones available for all ops
    if len(tq_ops) > len(tq_zones):
        raise GateOpportunitiesError
    if len(sq_ops) > len(sq_zones) - 2 * len(tq_ops):
        # Because SQ zones are offsets of TQ zones, each tq op covers 2 sq zones
        raise GateOpportunitiesError
    # place the tq ops
    order = place_tq_ops(tq_ops_sorted, placed_qubits, order, tq_zones, sq_zones)
    # run a check to avoid unnecessary swaps
    prev_state_inv = inverse(prev_state)
    for zone in tq_options:
        # enforce the relative ordering of qubits to prevent uneseccasry swaps
        # if the first qubit of a TQ gate was to the right of the second in prev_state
        # then there has been an unnecessary swap
        q0s = order[zone]
        q1s = order[zone + 1]
        if prev_state_inv[q0s] > prev_state_inv[q1s]:
            swapped = order[zone]
            order[zone] = order[zone + 1]
            order[zone + 1] = swapped

    # place the sq ops
    for op in sq_ops:
        q1 = op[0]
        # check to make sure that the qubits have not already been placed
        if q1 in placed_qubits:
            raise InvalidParallelOpsError(q1)

        prev_index = prev_state.index(q1)
        nearest_sq_zone = nearest(prev_index, sq_zones)
        order[nearest_sq_zone] = q1
        sq_zones.discard(nearest_sq_zone)
        placed_qubits.add(q1)

    # fill in the rest of the slots in the order with the inactive qubits
    for i in range(num_qubits):
        if i not in placed_qubits:
            prev_index = prev_state.index(i)
            nearest_sq_zone = nearest(prev_index, sq_zones)
            order[nearest_sq_zone] = i
            sq_zones.discard(nearest_sq_zone)

    if placement_check(ops, tq_options, sq_options, order):
        return order

    raise PlacementCheckError



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from typing import TYPE_CHECKING

from .placement import optimized_place
from .routing import transport_cost
from .sharding.shards2ops import parse_shards_naive

if TYPE_CHECKING:
    from .machine import Machine
    from .sharding.shard import Cost, Ordering, Shard, ShardLayer


def place_and_route(
    shards: list["Shard"],
    machine: "Machine | None" = None,
) -> list[tuple["Ordering", "ShardLayer", "Cost"]]:
    """Get all the routing info needed for PHIR generation."""
    shard_set = set(shards)
    circuit_rep, shard_layers = parse_shards_naive(shard_set)
    if machine:
        initial_order = list(range(machine.size))
    layer_num = 0
    orders: list[Ordering] = []
    layer_costs: list[Cost] = []
    net_cost: float = 0.0
    if machine:
        for layer in circuit_rep:
            order = optimized_place(
                layer,
                machine.tq_options,
                machine.sq_options,
                machine.size,
                initial_order,
            )
            orders.append(order)
            cost = transport_cost(initial_order, order, machine.qb_swap_time)
            layer_num += 1
            initial_order = order
            layer_costs.append(cost)
            net_cost += cost
    else:
        # If no machine object specified,
        # generic lists of qubits with no placement and no routing costs,
        # only the shards

        # If needed later, write a helper to find the number
        # of qubits needed in the circuit
        n = len(circuit_rep)
        orders = [[]] * n
        layer_costs = [0] * n

    # don't need a custom error for this, "strict" parameter will throw error if needed
    return list(zip(orders, shard_layers, layer_costs, strict=True))



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# Tests for qubit routing

import pytest

from pytket.phir.machine import Machine, MachineTimings
from pytket.phir.placement import (
    GateOpportunitiesError,
    InvalidParallelOpsError,
    place,
    placement_check,
)
from pytket.phir.qtm_machine import QTM_DEFAULT_GATESET

m = Machine(4, QTM_DEFAULT_GATESET, {1}, MachineTimings(10, 2, 2, 1))
m2 = Machine(6, QTM_DEFAULT_GATESET, {1, 3}, MachineTimings(10, 2, 2, 1))
m3 = Machine(8, QTM_DEFAULT_GATESET, {0, 6}, MachineTimings(10, 2, 2, 1))


def test_placement_check() -> None:
    """Test placement check."""
    # simple tq check
    ops = [[1, 2]]
    state = [0, 1, 2, 3]
    assert placement_check(ops, m.tq_options, m.sq_options, state)

    ops = [[2, 1]]
    state = [0, 1, 2, 3]
    assert placement_check(ops, m.tq_options, m.sq_options, state)

    # simple sq check
    ops = [[1], [2]]
    state = [0, 1, 2, 3]
    assert placement_check(ops, m.tq_options, m.sq_options, state)

    # combined sq/tq check
    ops = [[1, 2], [3], [4]]
    state = [0, 1, 2, 3, 4, 5]
    assert placement_check(ops, m2.tq_options, m2.sq_options, state)

    ops = [[2, 1], [3], [4]]
    state = [0, 1, 2, 3, 4, 5]
    assert placement_check(ops, m2.tq_options, m2.sq_options, state)

    # failing tests
    ops = [[0], [5]]
    state = [0, 1, 2, 3, 4, 5]
    assert not placement_check(ops, m2.tq_options, m2.sq_options, state)

    ops = [[1, 3], [2, 4]]
    state = [0, 1, 2, 3, 4, 5]
    assert not placement_check(ops, m2.tq_options, m2.sq_options, state)


def test_place() -> None:
    """Test place."""
    # one tq
    ops = [[0, 3]]
    tq_options = {1}
    sq_options = {1, 2}
    trap_size = 4
    expected = [1, 0, 3, 2]
    assert place(ops, tq_options, sq_options, trap_size) == expected

    # one tq two sq
    ops = [[0, 3], [1], [2]]
    tq_options = {1}
    sq_options = {0, 1, 2, 3}
    trap_size = 4
    expected = [1, 0, 3, 2]
    assert place(ops, tq_options, sq_options, trap_size) == expected

    # two tq
    ops = [[0, 5], [1, 4]]
    tq_options = {1, 3}
    sq_options = {0, 1, 2, 3, 4, 5}
    trap_size = 6
    expected = [2, 1, 4, 0, 5, 3]
    assert place(ops, tq_options, sq_options, trap_size) == expected

    # two tq two sq
    ops = [[0, 5], [1, 4], [2], [3]]
    tq_options = {1, 3}
    sq_options = {0, 1, 2, 3, 4, 5}
    trap_size = 6
    expected = [2, 1, 4, 0, 5, 3]
    assert place(ops, tq_options, sq_options, trap_size) == expected

    # real slice?
    ops = [
        [1, 5],
        [7, 15],
        [9, 11],
        [4, 28],
        [17, 23],
        [2],
        [10],
        [13],
        [16],
        [21],
        [22],
        [25],
    ]
    tq_options = {4, 8, 12, 16, 20, 24, 28}
    sq_options = set(range(32))
    trap_size = 32
    expected = [
        2,
        10,
        13,
        16,
        1,
        5,
        21,
        22,
        9,
        11,
        25,
        0,
        7,
        15,
        3,
        6,
        4,
        28,
        8,
        12,
        17,
        23,
        14,
        18,
        19,
        20,
        24,
        26,
        27,
        29,
        30,
        31,
    ]
    assert place(ops, tq_options, sq_options, trap_size) == expected

    # placement error
    ops = [[0, 1], [2]]
    tq_options = {0}
    sq_options = {0, 1}
    trap_size = 2
    with pytest.raises(GateOpportunitiesError):
        place(ops, tq_options, sq_options, trap_size)

    # op error
    ops = [[1], [1]]
    tq_options = {1}
    sq_options = {0, 1}
    trap_size = 2
    with pytest.raises(InvalidParallelOpsError):
        place(ops, tq_options, sq_options, trap_size)


test_placement_check()
test_place()



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here.
import datetime
import pathlib
import sys
from importlib.metadata import version as get_version

sys.path.insert(0, pathlib.Path(__file__).parents[2].resolve().as_posix())

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "pytket-phir"
project_copyright = f"2023-{datetime.date.today().year}, Quantinuum LLC"
author = "Quantinuum LLC"
release: str = ".".join(get_version("pytket-phir").split(".")[:2])

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
]

exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"



##############################################################################
#
# Copyright (c) 2024 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"

import base64
import hashlib
import json
import logging

import pytest

from pytket.circuit import Circuit, Qubit
from pytket.phir.api import pytket_to_phir, qasm_to_phir
from pytket.phir.qtm_machine import QtmMachine
from pytket.qasm.qasm import QASMUnsupportedError
from pytket.wasm.wasm import WasmModuleHandler

from .test_utils import WatFile, get_wat_as_wasm_bytes

logger = logging.getLogger(__name__)


def test_qasm_to_phir_with_wasm() -> None:
    """Test the qasm string entrypoint works with WASM."""
    qasm = """
    OPENQASM 2.0;
    include "qelib1.inc";

    qreg q[2];
    h q;
    ZZ q[1],q[0];
    creg cr[3];
    creg cs[3];
    creg co[3];
    measure q[0]->cr[0];
    measure q[1]->cr[1];

    cs = cr;
    co = add(cr, cs);
    """

    wasm_bytes = get_wat_as_wasm_bytes(WatFile.add)

    wasm_uid = hashlib.sha256(base64.b64encode(wasm_bytes)).hexdigest()

    phir_str = qasm_to_phir(qasm, QtmMachine.H1, wasm_bytes=wasm_bytes)
    phir = json.loads(phir_str)

    expected_metadata = {"ff_object": (f"WASM module uid: {wasm_uid}")}

    assert phir["ops"][21] == {
        "metadata": expected_metadata,
        "cop": "ffcall",
        "function": "add",
        "args": ["cr", "cs"],
        "returns": ["co"],
    }


def test_qasm_wasm_unsupported_reg_len() -> None:
    """Test that qasm containing calls to WASM with more than 32-bits fails."""
    qasm = """
    OPENQASM 2.0;
    include "qelib1.inc";

    creg cr[33];

    cr = add(cr, cr);
    """

    wasm_bytes = get_wat_as_wasm_bytes(WatFile.add)

    with pytest.raises(
        QASMUnsupportedError,
        match="limited to at most 32-bit|try setting the `maxwidth` parameter",
    ):
        qasm_to_phir(qasm, QtmMachine.H1, wasm_bytes=wasm_bytes)


def test_pytket_with_wasm() -> None:
    """Test whether pytket works with WASM."""
    w = WasmModuleHandler(get_wat_as_wasm_bytes(WatFile.testfile))

    c = Circuit()
    c0 = c.add_c_register("c0", 3)
    c1 = c.add_c_register("c1", 4)
    c2 = c.add_c_register("c2", 5)

    c.add_wasm_to_reg("multi", w, [c0, c1], [c2])
    c.add_wasm_to_reg("add_one", w, [c2], [c2])
    c.add_wasm_to_reg("no_return", w, [c2], [])
    c.add_wasm_to_reg("no_parameters", w, [], [c2])

    c.add_wasm_to_reg("add_one", w, [c0], [c0], condition=c1[0])

    phir_str = pytket_to_phir(c, QtmMachine.H1)

    phir = json.loads(phir_str)

    expected_metadata = {"ff_object": (f"WASM module uid: {w!s}")}

    assert phir["ops"][4] == {
        "metadata": expected_metadata,
        "cop": "ffcall",
        "function": "multi",
        "args": ["c0", "c1"],
        "returns": ["c2"],
    }
    assert phir["ops"][7] == {
        "metadata": expected_metadata,
        "cop": "ffcall",
        "function": "add_one",
        "args": ["c2"],
        "returns": ["c2"],
    }
    assert phir["ops"][8] == {
        "//": "IF ([c1[0]] == 1) THEN WASM_function='add_one' args=['c0'] returns=['c0'];"  # noqa: E501
    }
    assert phir["ops"][9] == {
        "block": "if",
        "condition": {"cop": "==", "args": [["c1", 0], 1]},
        "true_branch": [
            {
                "metadata": expected_metadata,
                "cop": "ffcall",
                "returns": ["c0"],
                "function": "add_one",
                "args": ["c0"],
            }
        ],
    }
    assert phir["ops"][12] == {
        "metadata": expected_metadata,
        "cop": "ffcall",
        "function": "no_return",
        "args": ["c2"],
    }
    assert phir["ops"][15] == {
        "metadata": expected_metadata,
        "cop": "ffcall",
        "function": "no_parameters",
        "args": [],
        "returns": ["c2"],
    }


def test_pytket_wasm_unsupported_reg_len() -> None:
    """Test that pytket circuit calling WASM with more than 32-bits fails."""
    w = WasmModuleHandler(get_wat_as_wasm_bytes(WatFile.testfile))

    c = Circuit(0, 33)
    c0 = c.add_c_register("c0", 33)

    with pytest.raises(ValueError, match="only registers of at most 32 bits"):
        c.add_wasm_to_reg("no_return", w, [c0], [])


def test_conditional_wasm() -> None:
    """From https://github.com/CQCL/pytket-phir/issues/156 ."""
    w = WasmModuleHandler(get_wat_as_wasm_bytes(WatFile.testfile))

    c = Circuit(1)
    areg = c.add_c_register("a", 2)
    breg = c.add_c_register("b", 1)
    c.H(0)
    c.Measure(Qubit(0), breg[0])
    c.add_wasm(
        funcname="add_one",
        filehandler=w,
        list_i=[1],
        list_o=[1],
        args=[areg[0], areg[1]],
        args_wasm=[0],
        condition_bits=[breg[0]],
        condition_value=1,
    )

    phir = json.loads(pytket_to_phir(c))

    assert phir["ops"][-2] == {
        "//": "IF ([b[0]] == 1) THEN WASM_function='add_one' args=['a'] returns=['a'];"
    }
    assert phir["ops"][-1]["true_branch"][0]["args"] == ["a"]



##############################################################################
#
# Copyright (c) 2023-2024 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"
# ruff: noqa: T201

import logging
from argparse import ArgumentParser
from importlib.metadata import version

from pecos.engines.hybrid_engine import HybridEngine
from pecos.foreign_objects.wasmtime import WasmtimeObj

from pytket.phir.phirgen import WORDSIZE
from pytket.qasm.qasm import (
    circuit_from_qasm,
    circuit_from_qasm_wasm,
)

from .api import pytket_to_phir
from .qtm_machine import QtmMachine


def main() -> None:
    """pytket-phir compiler CLI."""
    parser = ArgumentParser(
        prog="phirc",
        description="Emulates QASM program execution via PECOS",
    )
    parser.add_argument(
        "qasm_files", nargs="+", default=None, help="One or more QASM files to emulate"
    )
    parser.add_argument(
        "-w",
        "--wasm-file",
        default=None,
        help="Optional WASM file for use by the QASM programs",
    )
    parser.add_argument(
        "-m",
        "--machine",
        choices=["H1"],
        default="H1",
        help="Machine name, H1 by default",
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument(
        "--version",
        action="version",
        version=version("pytket-phir"),
    )
    args = parser.parse_args()

    for file in args.qasm_files:
        print(f"Processing {file}")
        circuit = None
        if args.wasm_file:
            print(f"Including WASM from file {args.wasm_file}")
            circuit = circuit_from_qasm_wasm(file, args.wasm_file, maxwidth=WORDSIZE)
            wasm_pecos_obj = WasmtimeObj(args.wasm_file)
        else:
            circuit = circuit_from_qasm(file, maxwidth=WORDSIZE)

        match args.machine:
            case "H1":
                machine = QtmMachine.H1
            case _:
                raise NotImplementedError

        if args.verbose:
            logging.basicConfig(level=logging.INFO)
        phir = pytket_to_phir(circuit, machine)

        print("\nPECOS results:")
        print(
            HybridEngine(qsim="state-vector").run(
                program=phir,
                shots=10,
                foreign_object=wasm_pecos_obj if args.wasm_file else None,  # type: ignore[arg-type]
            )
        )



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pytket.circuit import OpType


@dataclass
class MachineTimings:
    """Gate times for a machine.

    tq_time: time for a two qubit gate
    sq_time: time for a single qubit gate
    qb_swap_time: time it takes to swap to qubits
    meas_prep_time: time to arrange qubits for measurement
    """

    tq_time: float
    sq_time: float
    qb_swap_time: float
    meas_prep_time: float


class Machine:
    """A machine info class for testing."""

    def __init__(
        self,
        size: int,
        gateset: "set[OpType]",
        tq_options: set[int],
        timings: MachineTimings,
    ):
        """Create Machine object.

        Args:
            size: number of qubits/slots
            gateset: set of supported gates
            tq_options: options for where to perform tq gates
            timings: gate times
        """
        self.size = size
        self.gateset = gateset
        self.tq_options = tq_options
        self.sq_options: set[int] = set()
        self.tq_time = timings.tq_time
        self.sq_time = timings.sq_time
        self.qb_swap_time = timings.qb_swap_time
        self.meas_prep_time = timings.meas_prep_time

        for i in self.tq_options:
            self.sq_options.add(i)
            self.sq_options.add(i + 1)



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from typing import TYPE_CHECKING

from pytket.passes import AutoRebase, DecomposeBoxes
from pytket.phir.qtm_machine import QTM_DEFAULT_GATESET, QTM_MACHINES_MAP, QtmMachine

if TYPE_CHECKING:
    from pytket.circuit import Circuit


def rebase_to_qtm_machine(circuit: "Circuit", qtm_machine: QtmMachine) -> "Circuit":
    """Rebases a circuit's gate to the gate set appropriate for the given machine."""
    machine = QTM_MACHINES_MAP.get(qtm_machine)
    gateset = QTM_DEFAULT_GATESET if machine is None else machine.gateset
    c = circuit.copy()
    DecomposeBoxes().apply(c)
    AutoRebase(gateset, allow_swaps=True).apply(c)
    return c



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from pytket.circuit import Conditional, Op, OpType
from pytket.phir.sharding.sharder import Sharder

from .test_utils import QasmFile, get_qasm_as_circuit


class TestSharder:
    def test_shard_hashing(self) -> None:
        circuit = get_qasm_as_circuit(QasmFile.baby)
        shards = Sharder(circuit).shard()

        shard_set = set(shards)
        assert len(shard_set) == 3

        first_shard = next(iter(shard_set))
        shard_set.add(first_shard)
        assert len(shard_set) == 3

    def test_should_op_create_shard(self) -> None:
        expected_true: list[Op] = [
            Op.create(OpType.Measure),
            Op.create(OpType.Reset),
            Op.create(OpType.CX),
            Op.create(OpType.Barrier),
        ]
        expected_false: list[Op] = [
            Op.create(OpType.U1, 0.32),
            Op.create(OpType.H),
            Op.create(OpType.Z),
        ]

        for op in expected_true:
            assert Sharder.should_op_create_shard(op)
        for op in expected_false:
            assert not Sharder.should_op_create_shard(op)

    def test_with_baby_circuit(self) -> None:
        circuit = get_qasm_as_circuit(QasmFile.baby)
        shards = Sharder(circuit).shard()

        assert len(shards) == 3

        assert shards[0].primary_command.op.type == OpType.CX
        assert len(shards[0].primary_command.qubits) == 2
        assert not shards[0].primary_command.bits
        assert len(shards[0].sub_commands) == 2
        sub_commands = list(shards[0].sub_commands.items())
        assert sub_commands[0][1][0].op.type == OpType.H
        assert not shards[0].depends_upon

        assert shards[1].primary_command.op.type == OpType.Measure
        assert not shards[1].sub_commands
        assert shards[1].depends_upon == {shards[0].ID}

        assert shards[2].primary_command.op.type == OpType.Measure
        assert not shards[2].sub_commands
        assert shards[2].depends_upon == {shards[0].ID}

    def test_rollup_behavior(self) -> None:
        circuit = get_qasm_as_circuit(QasmFile.baby_with_rollup)
        shards = Sharder(circuit).shard()

        assert len(shards) == 5

        # Shard 0: CX gate, with 2 hadamard subcommands
        assert shards[0].primary_command.op.type == OpType.CX
        assert len(shards[0].primary_command.qubits) == 2
        assert not shards[0].primary_command.bits
        assert len(shards[0].sub_commands) == 2
        sub_commands = list(shards[0].sub_commands.items())
        assert sub_commands[0][1][0].op.type == OpType.H
        assert not shards[0].depends_upon

        # Shard 1: Measure q[0] -> c[0] (TKET splits q->c)
        assert shards[1].primary_command.op.type == OpType.Measure
        assert not shards[1].sub_commands
        assert shards[1].depends_upon == {shards[0].ID}

        # shard 2: Measure q[1] -> c[1]
        assert shards[2].primary_command.op.type == OpType.Measure
        assert not shards[2].sub_commands
        assert shards[2].depends_upon == {shards[0].ID}

        # Shard 3: Rollup barrier for a lingering hadamard on q[0]
        assert shards[3].primary_command.op.type == OpType.Barrier
        assert len(shards[3].sub_commands) == 1
        assert shards[3].depends_upon == {shards[1].ID}

        # Shard 4: Rollup barrier for second lingering hadamard on q[1]
        assert shards[4].primary_command.op.type == OpType.Barrier
        assert len(shards[4].sub_commands) == 1
        assert shards[4].depends_upon == {shards[2].ID}

    def test_simple_conditional(self) -> None:
        circuit = get_qasm_as_circuit(QasmFile.simple_cond)
        shards = Sharder(circuit).shard()

        assert len(shards) == 5

        # shard 0: [h q;] measure q->c;
        assert shards[0].primary_command.op.type == OpType.Measure
        assert shards[0].qubits_used == {circuit.qubits[0]}
        assert shards[0].bits_written == {circuit.bits[0]}
        assert not shards[0].depends_upon
        assert len(shards[0].sub_commands.items()) == 1
        s0_qubit, s0_sub_cmds = next(iter(shards[0].sub_commands.items()))
        assert s0_qubit == circuit.qubits[0]
        assert s0_sub_cmds[0].op.type == OpType.H

        # shard 1: reset q;
        assert shards[1].primary_command.op.type == OpType.Reset
        assert len(shards[1].sub_commands.items()) == 0
        assert shards[1].qubits_used == {circuit.qubits[0]}
        assert shards[1].depends_upon == {shards[0].ID}
        assert not shards[1].bits_written
        assert not shards[1].bits_read

        # shard 2: if (c==1) z=1;
        assert isinstance(shards[2].primary_command.op, Conditional)
        assert shards[2].primary_command.op.op.type == OpType.SetBits
        assert not shards[2].sub_commands
        assert not shards[2].qubits_used
        assert shards[2].bits_written == {circuit.bits[1]}
        assert shards[2].bits_read == {circuit.bits[0], circuit.bits[1]}
        assert shards[2].depends_upon == {shards[0].ID}

        # shard 3: [if (c==1) h q;]
        assert isinstance(shards[3].primary_command.op, Conditional)
        assert shards[3].primary_command.op.op.type == OpType.H
        assert not shards[3].sub_commands
        assert shards[3].qubits_used == {circuit.qubits[0]}
        assert shards[3].bits_read == {circuit.bits[0]}
        assert shards[3].depends_upon == {shards[0].ID, shards[1].ID}

        # shard 4: measure q->c;
        assert shards[4].primary_command.op.type == OpType.Measure
        assert not shards[4].sub_commands
        assert shards[4].qubits_used == {circuit.qubits[0]}
        assert shards[4].bits_written == {circuit.bits[0]}
        assert shards[4].depends_upon == {shards[0].ID, shards[3].ID}

    def test_complex_barriers(self) -> None:  # noqa: PLR0915
        circuit = get_qasm_as_circuit(QasmFile.barrier_complex)
        shards = Sharder(circuit).shard()

        assert len(shards) == 7

        # shard 0: [], c[3] = 1
        assert shards[0].primary_command.op.type == OpType.SetBits
        assert len(shards[0].sub_commands.items()) == 0
        assert not shards[0].qubits_used
        assert shards[0].bits_written == {circuit.bits[3]}
        assert shards[0].bits_read == {circuit.bits[3]}  # bits written are always read
        assert not shards[0].depends_upon

        # shard 1: [h q[0]; h q[1];] barrier q[0], q[1], c[3];
        assert shards[1].primary_command.op.type == OpType.Barrier
        assert len(shards[1].sub_commands.items()) == 2
        shard_1_q0_cmds = shards[1].sub_commands[circuit.qubits[0]]
        assert len(shard_1_q0_cmds) == 1
        assert shard_1_q0_cmds[0].op.type == OpType.H
        assert shard_1_q0_cmds[0].qubits == [circuit.qubits[0]]
        shard_1_q1_cmds = shards[1].sub_commands[circuit.qubits[1]]
        assert len(shard_1_q1_cmds) == 1
        assert shard_1_q1_cmds[0].op.type == OpType.H
        assert shard_1_q1_cmds[0].qubits == [circuit.qubits[1]]
        assert shards[1].qubits_used == {circuit.qubits[0], circuit.qubits[1]}
        assert shards[1].bits_written == {circuit.bits[3]}
        assert shards[1].bits_read == {circuit.bits[3]}
        assert shards[1].depends_upon == {shards[0].ID}

        # shard 2: [] CX q[0], q[1];
        assert shards[2].primary_command.op.type == OpType.CX
        assert len(shards[2].sub_commands.items()) == 0
        assert shards[2].qubits_used == {circuit.qubits[0], circuit.qubits[1]}
        assert not shards[2].bits_written
        assert not shards[2].bits_read
        assert shards[2].depends_upon == {shards[1].ID}

        # shard 3: measure q[0]->c[0];
        assert shards[3].primary_command.op.type == OpType.Measure
        assert len(shards[3].sub_commands.items()) == 0
        assert shards[3].qubits_used == {circuit.qubits[0]}
        assert shards[3].bits_written == {circuit.bits[0]}
        assert shards[3].bits_read == {circuit.bits[0]}
        assert shards[3].depends_upon == {shards[2].ID}

        # shard 4: [H q[3];, H q[3];, X q[3];]] barrier q[2], q[3];
        assert shards[4].primary_command.op.type == OpType.Barrier
        assert len(shards[4].sub_commands.items()) == 2
        shard_4_q2_cmds = shards[4].sub_commands[circuit.qubits[2]]
        assert len(shard_4_q2_cmds) == 2
        assert shard_4_q2_cmds[0].op.type == OpType.H
        assert shard_4_q2_cmds[0].qubits == [circuit.qubits[2]]
        assert shard_4_q2_cmds[1].op.type == OpType.H
        assert shard_4_q2_cmds[1].qubits == [circuit.qubits[2]]
        shard_4_q3_cmds = shards[4].sub_commands[circuit.qubits[3]]
        assert len(shard_4_q3_cmds) == 3
        assert shard_4_q3_cmds[0].op.type == OpType.H
        assert shard_4_q3_cmds[0].qubits == [circuit.qubits[3]]
        assert shard_4_q3_cmds[1].op.type == OpType.H
        assert shard_4_q3_cmds[1].qubits == [circuit.qubits[3]]
        assert shard_4_q3_cmds[2].op.type == OpType.X
        assert shard_4_q3_cmds[2].qubits == [circuit.qubits[3]]
        assert shards[4].qubits_used == {circuit.qubits[2], circuit.qubits[3]}
        assert not shards[4].bits_written
        assert not shards[4].bits_read
        assert not shards[4].depends_upon

        # shard 5: [] CX q[2], q[3];
        assert shards[5].primary_command.op.type == OpType.CX
        assert len(shards[5].sub_commands.items()) == 0
        assert shards[5].qubits_used == {circuit.qubits[2], circuit.qubits[3]}
        assert not shards[5].bits_written
        assert not shards[5].bits_read
        assert shards[5].depends_upon == {shards[4].ID}

        # shard 6: measure q[2]->c[2];
        assert shards[6].primary_command.op.type == OpType.Measure
        assert len(shards[6].sub_commands.items()) == 0
        assert shards[6].qubits_used == {circuit.qubits[2]}
        assert shards[6].bits_written == {circuit.bits[2]}
        assert shards[6].bits_read == {circuit.bits[2]}
        assert shards[6].depends_upon == {shards[5].ID}

    def test_classical_hazards(self) -> None:
        circuit = get_qasm_as_circuit(QasmFile.classical_hazards)
        shards = Sharder(circuit).shard()

        assert len(shards) == 5

        # shard 0: [h q[0];] measure q[0]->c[0];
        assert shards[0].primary_command.op.type == OpType.Measure
        assert len(shards[0].sub_commands.items()) == 1
        assert shards[0].qubits_used == {circuit.qubits[0]}
        assert shards[0].bits_written == {circuit.bits[0]}
        assert shards[0].bits_read == {circuit.bits[0]}
        assert not shards[0].depends_upon

        # shard 1: [H q[1];] measure q[1]->c[2];
        # NOTE: pytket reorganizes circuits to be efficiently ordered
        assert shards[1].primary_command.op.type == OpType.Measure
        assert len(shards[1].sub_commands.items()) == 1
        assert shards[1].qubits_used == {circuit.qubits[1]}
        assert shards[1].bits_written == {circuit.bits[2]}
        assert shards[1].bits_read == {circuit.bits[2]}
        assert not shards[1].depends_upon

        # shard 2: [] if(c[0]==1) c[1]=1;
        assert shards[2].primary_command.op.type == OpType.Conditional
        assert not shards[2].sub_commands
        assert not shards[2].qubits_used
        assert shards[2].bits_written == {circuit.bits[1]}
        assert shards[2].bits_read == {circuit.bits[1], circuit.bits[0]}
        assert shards[2].depends_upon == {shards[0].ID}

        # shard 3: [] c[0]=0;
        assert shards[3].primary_command.op.type == OpType.SetBits
        assert not shards[2].sub_commands
        assert not shards[3].qubits_used
        assert shards[3].bits_written == {circuit.bits[0]}
        assert shards[3].bits_read == {circuit.bits[0]}
        assert shards[3].depends_upon == {shards[0].ID, shards[2].ID}

        # shard 4: [] if(c[2]==1) c[0]=1;
        assert shards[4].primary_command.op.type == OpType.Conditional
        assert not shards[4].sub_commands
        assert not shards[4].qubits_used
        assert shards[4].bits_written == {circuit.bits[0]}
        assert shards[4].bits_read == {circuit.bits[0], circuit.bits[2]}
        assert shards[4].depends_upon == {shards[1].ID, shards[3].ID}

    def test_with_big_gate(self) -> None:
        circuit = get_qasm_as_circuit(QasmFile.big_gate)
        shards = Sharder(circuit).shard()

        assert len(shards) == 2

        # shard 0: [h q[0]; h q[1]; h q[2];] c4x q[0],q[1],q[2],q[3];
        assert shards[0].primary_command.op.type == OpType.CnX
        assert len(shards[0].sub_commands) == 3
        assert shards[0].qubits_used == {
            circuit.qubits[0],
            circuit.qubits[1],
            circuit.qubits[2],
            circuit.qubits[3],
        }
        assert not shards[0].bits_written

        # shard 1: [] measure q[3]->[c0]
        assert shards[1].primary_command.op.type == OpType.Measure
        assert not shards[1].sub_commands
        assert shards[1].qubits_used == {circuit.qubits[3]}
        assert shards[1].bits_written == {circuit.bits[0]}

    def test_classical_ordering_breaking_circuit(self) -> None:
        circuit = get_qasm_as_circuit(QasmFile.classical_ordering)
        shards = Sharder(circuit).shard()

        assert len(shards) == 4

        # shard 0
        assert shards[0].primary_command.op.type == OpType.SetBits
        assert len(shards[0].sub_commands.items()) == 0
        assert not shards[0].qubits_used
        assert shards[0].bits_written == {
            circuit.bits[0],
            circuit.bits[1],
            circuit.bits[2],
            circuit.bits[3],
        }
        assert shards[0].bits_read == {
            circuit.bits[0],
            circuit.bits[1],
            circuit.bits[2],
            circuit.bits[3],
        }
        assert not shards[0].depends_upon

        # shard 1
        assert shards[1].primary_command.op.type == OpType.CopyBits
        assert len(shards[1].sub_commands.items()) == 0
        assert not shards[1].qubits_used
        assert shards[1].bits_written == {
            circuit.bits[4],
            circuit.bits[5],
            circuit.bits[6],
            circuit.bits[7],
        }
        assert shards[1].bits_read == {
            circuit.bits[0],
            circuit.bits[1],
            circuit.bits[2],
            circuit.bits[3],
            circuit.bits[4],
            circuit.bits[5],
            circuit.bits[6],
            circuit.bits[7],
        }
        assert shards[1].depends_upon == {shards[0].ID}

        # shard 2
        assert shards[2].primary_command.op.type == OpType.ClassicalExpBox
        assert len(shards[2].sub_commands.items()) == 0
        assert not shards[2].qubits_used
        assert shards[2].bits_written == {
            circuit.bits[8],
            circuit.bits[9],
            circuit.bits[10],
            circuit.bits[11],
        }
        assert shards[2].bits_read == {
            circuit.bits[0],
            circuit.bits[1],
            circuit.bits[2],
            circuit.bits[3],
            circuit.bits[4],
            circuit.bits[5],
            circuit.bits[6],
            circuit.bits[7],
            circuit.bits[8],
            circuit.bits[9],
            circuit.bits[10],
            circuit.bits[11],
        }
        assert shards[2].depends_upon == {shards[0].ID, shards[1].ID}

        # shard 2
        assert shards[3].primary_command.op.type == OpType.ClassicalExpBox
        assert len(shards[3].sub_commands.items()) == 0
        assert not shards[3].qubits_used
        assert shards[3].bits_written == {
            circuit.bits[0],
            circuit.bits[1],
            circuit.bits[2],
            circuit.bits[3],
        }
        assert shards[3].bits_read == {
            circuit.bits[0],
            circuit.bits[1],
            circuit.bits[2],
            circuit.bits[3],
        }
        assert shards[3].depends_upon == {shards[0].ID, shards[2].ID}
